# Python standard lib
import os
from concurrent.futures import ThreadPoolExecutor

# Third party
import cv2 as cv
import geopandas as gpd
import matplotlib.pyplot as plt
import numpy as np
from pyproj import CRS, Transformer
import rasterio
from rasterio.features import geometry_mask
from rasterio.warp import calculate_default_transform, reproject, Resampling
from osgeo import gdal, osr
from shapely.geometry import Polygon, mapping, MultiPoint
from sklearn.neighbors import NearestNeighbors
from spectral import envi
import spectral as sp
import h5py
from scipy.spatial.transform import Rotation as RotLib

# Lib modules
from gref4hsi.utils.colours import Image as Imcol

# ENVI datatype conversion dictionary
dtype_dict = {1:np.uint8,
             2:np.int16,
             3:np.int32,
             4:np.float32,
             5:np.float64,
             12:np.uint16,
             13:np.uint32,
             14:np.int64,
             15:np.uint64}

class GeoSpatialAbstractionHSI():
    def __init__(self, point_cloud, transect_string, config_crs):
        self.name = transect_string
        self.points_geocsc = point_cloud

        self.epsg_geocsc = config_crs.epsg_geocsc
        
        self.n_lines = point_cloud.shape[0]
        self.n_pixels = point_cloud.shape[1]
        self.n_bands = point_cloud.shape[1]

        # A clean way of doing things would be to define 
    def transform_geocentric_to_projected(self, config_crs):


        
        self.epsg_proj = config_crs.epsg_proj
        self.crs = 'EPSG:' + str(self.epsg_proj)
        
        

        self.points_proj  = np.zeros(self.points_geocsc.shape) # Remains same if it is local
        
        geocsc = CRS.from_epsg(self.epsg_geocsc)
        proj = CRS.from_epsg(self.epsg_proj)
        transformer = Transformer.from_crs(geocsc, proj)

        x_ecef = self.points_geocsc[:,:,0].reshape((-1, 1))
        y_ecef = self.points_geocsc[:, :, 1].reshape((-1, 1))
        z_ecef = self.points_geocsc[:, :, 2].reshape((-1, 1))

        

        (east, north, hei) = transformer.transform(xx=x_ecef, yy=y_ecef, zz=z_ecef)

        self.points_proj[:,:,0] = east.reshape((self.points_proj.shape[0], self.points_proj.shape[1]))
        self.points_proj[:, :, 1] = north.reshape((self.points_proj.shape[0], self.points_proj.shape[1]))
        self.points_proj[:, :, 2] = hei.reshape((self.points_proj.shape[0], self.points_proj.shape[1]))


        


    def footprint_to_shape_file(self, footprint_dir):
        """Uses the georeferenced data points (in projected form) to describe the footprint as a shape file

        :param footprint_dir: Where to put the shape files describing the footprint
        :type footprint_dir: string
        """
        self.edge_start = self.points_proj[0, :, 0:2].reshape((-1,2))
        self.edge_end = self.points_proj[-1, :, 0:2].reshape((-1,2))
        self.side_1 = self.points_proj[:, 0, 0:2].reshape((-1,2))
        self.side_2 = self.points_proj[:, -1, 0:2].reshape((-1,2))


        # Do it clockwise
        self.hull_line = np.concatenate((
            self.edge_start,
            self.side_2,
            np.flip(self.edge_end, axis=0),
            np.flip(self.side_1, axis = 0)

        ), axis = 0)

        # The swiped ground area is defined by the convex hull
        self.footprint_shp = Polygon(self.hull_line).convex_hull

        gdf = gpd.GeoDataFrame(geometry=[self.footprint_shp], crs=self.crs)

        shape_path = footprint_dir + self.name + '.shp'

        gdf.to_file(shape_path, driver='ESRI Shapefile')

    def resample_datacube(self, radiance_cube, wavelengths, fwhm, envi_cube_dir, rgb_composite_dir, config_ortho):
        """Resamples the radiance cube into a geographic grid based on the georeferencing

        :param radiance_cube: The data cube of radiance with the corresponding radiometric_unit of the data
        :type radiance_cube: Often an ndarray(n, m, k) where n-number of lines, m- number of pixels, and k-number of spectral bands
        :param wavelengths: The band's centre wavelengths
        :type wavelengths: ndarray(k, 1)
        :param fwhm: Full Width Half Maximum, descibing the band's widths (often in nanometers)
        :type fwhm: ndarray(k, 1)
        :param envi_cube_dir: Directory to write data cubes
        :type envi_cube_dir: string, path
        :param rgb_composite_dir: Directory to write data cubes
        :type rgb_composite_dir: string, path
        :param config_ortho: The relevant configurations for orthorectification
        :type config_ortho: Dictionary
        """



        
        n_bands = len(wavelengths)
        #
        n = radiance_cube.shape[0]
        m = radiance_cube.shape[1]
        k = radiance_cube.shape[2] # Number of bands

        bytes_per_entry = radiance_cube.itemsize
        chunk_size_GB = config_ortho.chunk_size_cube_GB


        # If chunking is to be applied, we can use square chunks. Round to the nearest thousand because of personal OCD
        chunk_square_length = np.sqrt((chunk_size_GB*1024**3) / (k*bytes_per_entry))
        self.chunk_square_length = int(np.round(chunk_square_length/1000)*1000)

        if self.chunk_square_length == 0:
            self.chunk_square_length = 1000

        self.chunk_area = self.chunk_square_length**2

        self.res = config_ortho.ground_resolution

        wl_red = config_ortho.wl_red
        wl_green = config_ortho.wl_green
        wl_blue = config_ortho.wl_blue

        raster_transform_method = config_ortho.raster_transform_method

        # Set nodata value for ortho-products
        nodata = config_ortho.nodata_value
        self.nodata = nodata

        rgb_composite_only = config_ortho.resample_rgb_only

        

        # Set custom RGB bands from *.ini file
        band_ind_R = np.argmin(np.abs(wl_red - wavelengths))
        band_ind_G = np.argmin(np.abs(wl_green - wavelengths))
        band_ind_B = np.argmin(np.abs(wl_blue - wavelengths))

        # To let ENVI pick up on which bands are used for red-green-blue vizualization
        self.default_bands_string = '{ '+' , '.join([str(band_ind_R), str(band_ind_G), str(band_ind_B)]) + ' }'

        # Some relevant metadata
        metadata_ENVI = {
            'description': 'Radiance converted, georeferenced data',
            'unit': config_ortho.radiometric_unit,
            'wavelength units': config_ortho.wavelength_unit,
            'sensor type': config_ortho.sensor_type,
            'default bands': self.default_bands_string,
            'interleave': config_ortho.interleave,
            'wavelengths': wavelengths
        }
        try:
            # If vector form
            if fwhm.any() == np.nan:
                pass
            else:
                metadata_ENVI['fwhm'] = fwhm
        except AttributeError:
            # If scalar
            if fwhm == np.nan:
                pass
            else:
                metadata_ENVI['fwhm'] = fwhm

            
        
        
        # Extract relevant info from hyp object
        datacube = radiance_cube[:, :, :].reshape((-1, n_bands))

        # RBG Composite
        rgb_cube = datacube[:, [band_ind_R, band_ind_G, band_ind_B]].reshape((-1, 3))

        # Horizontal coordinates of intersections in projected CRS
        coords = self.points_proj[:, :, 0:2].reshape((-1, 2))

        del radiance_cube
        
        # The raster can be rotated optimally (which saves loads of memory) for transects that are long compared to width. 
        # However, north-east oriented rasters is more supported by image visualization
        transform, height, width, indexes, suffix, mask_nn = GeoSpatialAbstractionHSI.cube_to_raster_grid(coords, raster_transform_method, resolution = self.res)

        # Make accessible as attribute because it can be to write ancillary data
        self.indexes = indexes.copy()
        self.transform = transform
        self.width = width
        self.height = height
        self.suffix = suffix

        # Create raster mask from the polygon describing the footprint (currently not used for anything)
        #geoms = [mapping(self.footprint_shp)]
        #mask = geometry_mask(geoms, out_shape=(height, width), transform=transform)

        mask = mask_nn.reshape((height, width))
        self.mask = mask

        # Build datacube
        if not rgb_composite_only:

            # For the later processing, storing the mapping from the rectified grid to the raw datacube makes sense:
            indexes_grid_unmasked = indexes.reshape((height, width))
            # Mask indexes
            indexes_grid_unmasked[mask == 1] = nodata

            # Make masked indices accessible as these allow orthorectification of ancilliary data
            self.index_grid_masked = indexes_grid_unmasked

            # To do the writing of the rasters we use a memory map, and writes chunks at a time.
            # For this writing process, we need to know all the below
            memmap_gen_params = {
                'indexes': indexes,
                'index_grid_masked': self.index_grid_masked,
                'mask': mask,
                'nodata': nodata,
                'height': height,
                'width': width,
                'datacube': datacube,
                'chunk_area': self.chunk_area
            }

            GeoSpatialAbstractionHSI.write_datacube_ENVI(memmap_gen_params, 
                                nodata, 
                                transform, 
                                datacube_path = envi_cube_dir + self.name + suffix, 
                                wavelengths=wavelengths,
                                fwhm=fwhm,
                                metadata = metadata_ENVI,
                                crs=self.crs,
                                interleave = config_ortho.interleave)

            

        # Resample RGB image data 
        ortho_rgb = rgb_cube[self.indexes, :].flatten()

        # Reshape image
        ortho_rgb = ortho_rgb.reshape((height, width, 3))
    
        # Mask RGB
        ortho_rgb[mask == 1, :] = nodata

        # Arange datacube or composite in rasterio-friendly structure
        ortho_rgb = np.transpose(ortho_rgb, axes = [2, 0, 1])
        
        # Write pseudo-RGB composite to composite folder ../GIS/RGBComposites
        with rasterio.open(rgb_composite_dir + self.name + suffix + '.tif', 'w', driver='GTiff',
                                height=height, width=width, count=3, dtype=ortho_rgb.dtype,
                                crs=self.crs, transform=transform, nodata=nodata) as dst:

            dst.write(ortho_rgb)
        
    @staticmethod
    def write_datacube_memmap(memmap_array, indexes, index_grid_masked, mask, nodata, height, width, datacube, chunk_area):  
        """Function for writing data cube arrays to memory maps, effectively avoiding memory issues.

        :param memmap_array: _description_
        :type memmap_array: _type_
        :param indexes: _description_
        :type indexes: _type_
        :param index_grid_masked: _description_
        :type index_grid_masked: _type_
        :param mask: _description_
        :type mask: _type_
        :param nodata: _description_
        :type nodata: _type_
        :param height: _description_
        :type height: _type_
        :param width: _description_
        :type width: _type_
        :param datacube: _description_
        :type datacube: _type_
        :param chunk_area: _description_
        :type chunk_area: _type_
        """

        # How the hell is this logical?
        n_bands = datacube.shape[1]

        # In grid form, identify whether partitioning is needed
        area_idx_grid = width*height

        if area_idx_grid > chunk_area:
            
            # Subdivide along horizontal axis:
            delta_width = int(chunk_area/height)
            # Number of slices
            num_delta_widths = int(width/delta_width) + 1
            # Iterate slices
            for i in range(num_delta_widths):
                # Last slice has different thickness
                if i != num_delta_widths-1:
                    sub_indices = index_grid_masked[:, i*delta_width:(i+1)*delta_width].reshape((-1,1))
                    dw = delta_width
                    
                else:
                    sub_indices = index_grid_masked[:, i*delta_width:width].reshape((-1,1))
                    dw = np.arange(i*delta_width, width).size
                
                # Only extraxt valid data:
                sub_ortho_cube = np.ones((sub_indices.shape[0]*sub_indices.shape[1], n_bands))*nodata
                
                # From the grid, the only valid pixels are
                valid_pixels = (sub_indices != nodata).reshape(-1)
                valid_pixels_raw = sub_indices[valid_pixels].reshape(-1)
                valid_pixels_geo = np.arange(sub_indices.size)[valid_pixels]

                sub_ortho_cube[valid_pixels_geo, :] = datacube[valid_pixels_raw, :]

                sub_ortho_cube = sub_ortho_cube.reshape((height, dw, n_bands))

                # Form to Rasterio friendly
                #sub_ortho_cube = np.transpose(sub_ortho_cube, axes=[2, 0, 1])

                memmap_array[:, i*delta_width:i*delta_width + dw, :] = sub_ortho_cube
            # Free memory
            del datacube
            del sub_ortho_cube
        else:
            # Resample
            ortho_datacube = datacube[indexes, :]

            # Free memory
            del datacube

            # Reshape to datacube form
            ortho_datacube = ortho_datacube.reshape((height, width, n_bands))

            # Mask
            ortho_datacube[mask == 1, :] = nodata

            # Form to Rasterio friendly
            #ortho_datacube = np.transpose(ortho_datacube, axes=[2, 0, 1])

            # Write to memory map
            memmap_array[:] = ortho_datacube

            # Free memory
            del ortho_datacube

        return  

    def resample_ancillary(self, h5_filename, anc_dir, anc_dict, interleave = 'bsq'):

        band_counter = 0
        band_names = []

        # Define one hyperspectral data to 
        with h5py.File(h5_filename, 'r', libver='latest') as f:
            for attribute_name, h5_hierarchy_item_path in anc_dict.items():
                if attribute_name != 'folder':
                    data = f[h5_hierarchy_item_path][()]


                    if data.ndim == 2:
                        if data.shape[1]  != self.n_pixels:
                            # For data of shape n_lines x j, for example camera positions, reshape to n_lines x n_pixels x j
                            j = data.shape[1]
                            data = np.einsum('ijk, ik -> ijk', np.ones((self.n_lines, self.n_pixels, j), dtype=np.float64), data)
                        else:
                            # For data with dimension n_lines x n_pixels, add third dimension
                            data = data.reshape((data.shape[0], data.shape[1], 1))

                    data = data.astype(dtype = np.float64)
                    
                    # For first layer
                    if band_counter == 0:
                        anc_data_array = data
                    # For other layers, concatenate with existing layers
                    else:
                        anc_data_array = np.concatenate((anc_data_array, data), axis = 2)

                    # Necessary to modify for data with multiple bands
                    k = data.shape[2]
                    # If data has more than one band:
                    if k > 1:
                        for i in range(k):
                            band_name_data = attribute_name + '_' + str(i)
                            band_names.append(band_name_data)
                    else:
                        band_name_data = attribute_name
                        band_names.append(band_name_data)
                    
                    band_counter += k
                    
                    
            
        
        
        metadata_anc = {
            'description': 'Ancillary data',
            'band names': '{ '+' , '.join(band_names) + ' }'
        }
        n_bands = band_counter

        # Resample
        anc_data_array = anc_data_array.reshape((-1, n_bands))
        
        # Grid of indices
        indexes_grid_unmasked = self.indexes.reshape((self.height, self.width))

        # Mask indexes
        indexes_grid_unmasked[self.mask == 1] = self.nodata

        # Make masked indices accessible as these allow orthorectification of ancilliary data
        self.index_grid_masked = indexes_grid_unmasked

        # Alternative 2
        memmap_gen_params = {
            'indexes': self.indexes,
            'index_grid_masked': self.index_grid_masked,
            'mask': self.mask,
            'nodata': self.nodata,
            'height': self.height,
            'width': self.width,
            'datacube': anc_data_array,
            'chunk_area': self.chunk_area
        }
        
        GeoSpatialAbstractionHSI.write_ancillary_ENVI_envi(nodata = self.nodata, 
                                                    transform = self.transform, 
                                                    crs = self.crs,
                                                    anc_path = anc_dir + self.name + self.suffix,
                                                    metadata = metadata_anc,
                                                    interleave=interleave,
                                                    memmap_gen_params=memmap_gen_params)

        

    @staticmethod
    def cube_to_raster_grid(coords, raster_transform_method, resolution):
        """Function that takes projected coordinates (e.g. UTM 32 east and north) of ray intersections and computes an image grid

        :param coords: _description_
        :type coords: _type_
        :param raster_transform_method: How the raster grid is calculated. "north_east" is standard and defines a rectangle along north/east. "minimal_rectangle" is memory optimal as it finds the smallest enclosing rectangle that wraps the points.
        :type raster_transform_method: _type_
        :param resolution: _description_
        :type resolution: _type_
        :return: _description_
        :rtype: _type_
        """


        if raster_transform_method == 'north_east':
            # Creates the minimal area (and thus memory) rectangle around points
            polygon = MultiPoint(coords).envelope

            # extract coordinates
            x, y = polygon.exterior.xy

            suffix = '_north_east'
            

            
        elif raster_transform_method == 'minimal_rectangle':
            
            # Creates the minimal area (and thus memory) rectangle around points
            polygon = MultiPoint(coords).minimum_rotated_rectangle

            # extract coordinates
            x, y = polygon.exterior.xy

            

            suffix = '_rotated'

        # Indices increase against the clock
        # ul means upper left corner of polygon, ll means lower left and so on
        idx_ul = 3
        idx_ur = 2
        idx_ll = 4

        x_ul = x[idx_ul]
        y_ul = y[idx_ul]

        x_ur = x[idx_ur]
        y_ur = y[idx_ur]

        x_ll = x[idx_ll]
        y_ll = y[idx_ll]

        # The vector from the upper-left corner aka origin to the upper-right equals lambda*e_basis_x
        e_basis_x = np.array([x_ur-x_ul, y_ur-y_ul]).reshape((2,1))
        w_transect = np.linalg.norm(e_basis_x)
        e_basis_x /= w_transect

        # The y basis vector is the vector to the other edge
        e_basis_y = np.array([x_ll-x_ul, y_ll-y_ul]).reshape((2,1))
        h_transect = np.linalg.norm(e_basis_y)
        e_basis_y /= h_transect
        e_basis_y *= -1 # Ensuring right handedness (image storage y direction is opposite)

        R = np.hstack((e_basis_x, e_basis_y)) # 2D rotation matrix by definition
        
        # Define origin/translation vector as the upper left corner
        o = np.array([x[idx_ul], y[idx_ul]]).reshape((2))

        # Transformation matrix rigid body 2D
        Trb = np.zeros((3,3))
        Trb[0:2,0:2] = R
        Trb[0:2,2] = o
        Trb[2,2] = 1

        # We operate with same resolution in X and Y. Note that for "minimal_rectangle" X, Y are NOT East and North. 
        s_x = resolution
        s_y = resolution

        S = np.diag(np.array([s_x, s_y, 1]))

        # Reflection matrix to account for opposite row/up direction
        Ref = np.diag(np.array([1, -1, 1]))

        # The affine transform is then expressed:
        Taff = Trb.dot(S).dot(Ref)

        # The metric width and length of the transect can give us the number of pixels
        width = int(np.ceil(w_transect/s_x))
        height = int(np.ceil(h_transect/s_y))

        # Rasterio operates with a conventional affine geotransform
        a, b, c, d, e, f = Taff[0,0], Taff[0,1], Taff[0,2], Taff[1,0], Taff[1,1], Taff[1,2]
        transform = rasterio.Affine(a, b, c, d, e, f)

        # Define local orthographic pixel grid. Pixel centers reside at half coordinates.
        xi, yi = np.meshgrid(np.arange(width) + 0.5, 
                                np.arange(height) + 0.5)
        # To get homogeneous vector (not an actual z coordinate)
        zi = np.ones(xi.shape)

        # Form homogeneous vectors (allows translation by multiplication)
        x_r = np.vstack((xi.flatten(), yi.flatten(), zi.flatten())).T

        # Map orthographic pixels to projected system
        x_p = np.matmul(Taff, x_r.T).T 

        # Make a point vector of grid points to be used in nearest neighbor
        xy = np.vstack((x_p[:,0].flatten(), x_p[:,1].flatten())).T
        
        # Define NN search tree from intersection points
        tree = NearestNeighbors(radius=resolution).fit(coords)

        # Calculate the nearest intersection point (in "coords") for each grid cell ("xy").
        
        # Here we only use one neighbor, and indexes is a vector of len(xy) where an element indexes(i)
        # says that the closest point to xy[i] is coords[indexes[i]]. Since coords is just a flattened/reshaped version of intersection points
        #, the data cube can be resampled by datacube_flat=datacube.reshape((dim1*dim2, dim3)) and 
        # geographic_datacube_flat = datacube.reshape((dim1_geo*dim2_geo, dim3)) so that geographic_datacube_flat = datacube_flat[indexes,:]
        n_neighbors = 1
        dist, indexes = tree.kneighbors(xy, n_neighbors)

        # We can mask the data by allowing points within a radius of 2x the resolution
        mask_nn = dist > 2*resolution
        
        return transform, height, width, indexes, suffix, mask_nn

    @staticmethod
    def write_datacube_ENVI(memmap_gen_params, nodata, transform, datacube_path, wavelengths, fwhm, metadata, interleave, crs):
        """_summary_

        :param ortho_datacube: _description_
        :type ortho_datacube: _type_
        :param nodata: _description_
        :type nodata: _type_
        :param transform: _description_
        :type transform: _type_
        :param datacube_path: _description_
        :type datacube_path: _type_
        :param wavelengths: _description_
        :type wavelengths: _type_
        :param fwhm: _description_
        :type fwhm: _type_
        :param envi_hdr_dict: _description_
        :type envi_hdr_dict: _type_
        """

        nx = memmap_gen_params['height']
        mx = memmap_gen_params['width']
        k = memmap_gen_params['datacube'].shape[1] # is collapsed datacube

        dtype_cube = memmap_gen_params['datacube'].dtype # is collapsed datacube

        # Make some simple modifications
        data_file_path = datacube_path + '.' + interleave
        header_file_path = datacube_path + '.hdr'
        

        if os.path.exists(data_file_path):
            os.remove(data_file_path)
            os.remove(header_file_path)

        
        # Create metadata for the output raster dataset
        crs_gdal = osr.SpatialReference()
        
        crs_gdal.ImportFromEPSG(int(crs.split(':')[1]))

        # Create 1x1x1 dummy file to exploit builtin driver
        with rasterio.open(data_file_path, 'w', driver='ENVI', height=1, width=1, count=1, crs=crs, dtype=dtype_cube, transform=transform, nodata=nodata) as dst:
            pass

        # Then remove 
        os.remove(data_file_path)

        header = sp.io.envi.read_envi_header(datacube_path + '.hdr') # Open for extraction
        
        # Since dummy image was made as 1x1x1 data cube
        metadata_dim = {
            'lines': nx,
            'samples': mx,
            'bands': k
        }

        # Write all meta data to header
        for meta_key, value in metadata_dim.items():
            header[meta_key] = value

        # Reshape the ortho_datacube to (nx, mx, k)
        #ortho_datacube_reshaped = ortho_datacube.transpose(1, 2, 0)

        # Create the output raster dataset
        # Create ENVI image without using a context manager
        dst = envi.create_image(datacube_path + '.hdr', interleave='bsq', metadata=header, force=True)

        mm = dst.open_memmap(writable=True)

        GeoSpatialAbstractionHSI.write_datacube_memmap(memmap_array=mm, **memmap_gen_params)
        
        header = sp.io.envi.read_envi_header(datacube_path + '.hdr') # Open for extraction
        header.pop('band names')

        # Nobody in the history of the world could have come up with a more annoying bug.
        # Apparently, the CRS string is written with white spaces by rasterio/GDAL, wheras it should have none.
        header['coordinate system string'] = '{' + ",".join(header['coordinate system string']) + '}'
        
        # Write all meta_data to header
        for meta_key, value in metadata.items():
            header[meta_key] = value

        sp.io.envi.write_envi_header(fileName=header_file_path, header_dict=header)

    @staticmethod
    def write_ancillary_ENVI(anc_data, nodata, transform, anc_path, metadata, interleave, crs):
        """_summary_

        :param anc_data: An ancilliary data cube 
        :type anc_data: j x m x n where j are the number of bands, and m, n are the raster dimensions 
        :param nodata: _description_
        :type nodata: _type_
        :param transform: _description_
        :type transform: _type_
        :param datacube_path: _description_
        :type datacube_path: _type_
        :param metadata: _description_
        :type metadata: _type_
        :param interleave: _description_
        :type interleave: _type_
        :param crs: _description_
        :type crs: _type_
        """
        nx = anc_data.shape[1]
        mx = anc_data.shape[2]
        k = anc_data.shape[0]

        # Make some simple modifications
        data_file_path = anc_path + '.' + interleave
        header_file_path = anc_path + '.hdr'

        if os.path.exists(data_file_path):
            os.remove(data_file_path)
            os.remove(header_file_path)

        # Hack to exploit rasterio's generation of map info

        # Clean the files generated by rasterio
        def write_band(args):
            band_data, index, dst = args
            dst.write(band_data, index + 1)

        # Assuming ortho_datacube is a 3D NumPy array with shape (k, nx, mx)
        with rasterio.open(data_file_path, 'w', driver='ENVI', height=nx, width=mx, count=k, crs=crs, dtype=anc_data.dtype, transform=transform, nodata=nodata) as dst:
            # Create a ThreadPoolExecutor with as many threads as bands
            with ThreadPoolExecutor(max_workers=k) as executor:
                # Use executor.map to parallelize the band writing process
                executor.map(write_band, [(band_data, i, dst) for i, band_data in enumerate(anc_data)])
            
        
        header = sp.io.envi.read_envi_header(anc_path + '.hdr') # Open for modif

        header.pop('band names') # Remove defaults

        for meta_key, value in metadata.items():
            header[meta_key] = value

        sp.io.envi.write_envi_header(fileName=header_file_path, header_dict=header)

    @staticmethod
    def write_ancillary_ENVI_envi(nodata, transform, anc_path, metadata, interleave, crs, memmap_gen_params):
        """_summary_

        :param anc_data: An ancilliary data cube 
        :type anc_data: j x m x n where j are the number of bands, and m, n are the raster dimensions 
        :param nodata: _description_
        :type nodata: _type_
        :param transform: _description_
        :type transform: _type_
        :param datacube_path: _description_
        :type datacube_path: _type_
        :param metadata: _description_
        :type metadata: _type_
        :param interleave: _description_
        :type interleave: _type_
        :param crs: _description_
        :type crs: _type_
        """
        nx = memmap_gen_params['height']
        mx = memmap_gen_params['width']
        k = memmap_gen_params['datacube'].shape[1] # is collapsed datacube

        dtype_cube = memmap_gen_params['datacube'].dtype # is collapsed datacube
        # Make some simple modifications
        data_file_path = anc_path + '.' + interleave
        header_file_path = anc_path + '.hdr'

        #if os.path.exists(header_file_path):
        #    return
        

        if os.path.exists(data_file_path):
            os.remove(data_file_path)
            os.remove(header_file_path)


        # Create metadata for the output raster dataset
        crs_gdal = osr.SpatialReference()
        
        crs_gdal.ImportFromEPSG(int(crs.split(':')[1]))

        # Create 1x1x1 dummy file to exploit builtin driver
        with rasterio.open(data_file_path, 'w', driver='ENVI', height=1, width=1, count=1, crs=crs, dtype=dtype_cube, transform=transform, nodata=nodata) as dst:
            pass

        # Then remove 
        os.remove(data_file_path)

        # Open for extraction
        header = sp.io.envi.read_envi_header(anc_path + '.hdr')

        # Since dummy image was made as 1x1x1 data cube
        metadata_dim = {
            'lines': nx,
            'samples': mx,
            'bands': k
        }

        # Write all dimensions to header
        for meta_key, value in metadata_dim.items():
            header[meta_key] = value
        
        # Create ENVI image without using a context manager
        dst = envi.create_image(anc_path + '.hdr', interleave='bsq', metadata=header, force=True)

        mm = dst.open_memmap(writable=True)

        # Write to the memory map
        GeoSpatialAbstractionHSI.write_datacube_memmap(memmap_array=mm, **memmap_gen_params)
        
        header = sp.io.envi.read_envi_header(anc_path + '.hdr') # Open for extraction
        header.pop('band names')

        # Nobody in the history of the world could have come up with a more annoying bug.
        # Apparently, the CRS string is written with white spaces by rasterio/GDAL, wheras it should have none.
        header['coordinate system string'] = '{' + ",".join(header['coordinate system string']) + '}'
        
        # Write all meta_data to header
        for meta_key, value in metadata.items():
            header[meta_key] = value

        sp.io.envi.write_envi_header(fileName=header_file_path, header_dict=header)



    @staticmethod
    def compare_hsi_composite_with_rgb_mosaic(hsi_composite_path, ref_ortho_reshaped_path):
        """Compares an HSI orthomosaic """
        
        

        
        # The RGB orthomosaic after reshaping (the reference)
        raster_rgb = gdal.Open(ref_ortho_reshaped_path, gdal.GA_Update)
        xoff1, a1, b1, yoff1, d1, e1 = raster_rgb.GetGeoTransform()  # This should be equal
        raster_rgb_array = np.array(raster_rgb.ReadAsArray())
        R = raster_rgb_array[0, :, :].reshape((raster_rgb_array.shape[1], raster_rgb_array.shape[2], 1))
        G = raster_rgb_array[1, :, :].reshape((raster_rgb_array.shape[1], raster_rgb_array.shape[2], 1))
        B = raster_rgb_array[2, :, :].reshape((raster_rgb_array.shape[1], raster_rgb_array.shape[2], 1))
        # del raster_array1
        ortho_rgb = np.concatenate((R, G, B), axis=2)
        rgb_image = Imcol(ortho_rgb)

        # The HSI composite raster (the match)
        raster_hsi = gdal.Open(hsi_composite_path)
        raster_hsi_array = np.array(raster_hsi.ReadAsArray())
        xoff2, a2, b2, yoff2, d2, e2 = raster_hsi.GetGeoTransform()
        transform_pixel_projected = raster_hsi.GetGeoTransform()
        R = raster_hsi_array[0, :, :].reshape((raster_hsi_array.shape[1], raster_hsi_array.shape[2], 1))
        G = raster_hsi_array[1, :, :].reshape((raster_hsi_array.shape[1], raster_hsi_array.shape[2], 1))
        B = raster_hsi_array[2, :, :].reshape((raster_hsi_array.shape[1], raster_hsi_array.shape[2], 1))
        ortho_hsi = np.concatenate((R, G, B), axis=2)

        # Some form of image processing

        no_data_value = raster_hsi.GetRasterBand(1).GetNoDataValue()

        ortho_valid = ortho_hsi[ortho_hsi[:,:,0] != no_data_value]

        # Iterate the colors
        for i in range(3):
            # Enhancing colors somewhat
            max_val = np.percentile(ortho_valid[:,i].reshape(-1), 95)
            min_val = 0
            
            mono_im = ortho_hsi[:,:,i]
            ortho_hsi[:,:,i] = (mono_im-min_val)/(max_val - min_val)

            # Remove out-of-bounds
            ortho_hsi[mono_im > 1, i] = 1
            ortho_hsi[mono_im < 0, i] = 0
        
        
        
        ortho_hsi = ortho_hsi * 255
        ortho_hsi[ortho_hsi == 0] = 255
        hsi_image = Imcol(ortho_hsi)

        # Adjust Clahe
        #hsi_image.clahe_adjustment()
        rgb_image.clahe_adjustment()

        # Radiance is equivalent to a linear "color space". We run it through an inverse gamma to match the RGB data
        hsi_image.to_luma(gamma=False, 
                          image_array = hsi_image.image_array, 
                          gamma_inverse=True, 
                          gamma_value=0.45)
        
        hsi_image.clahe_adjustment(is_luma = True)

        rgb_image.to_luma(gamma=False, image_array= rgb_image.clahe_adjusted)

        uv_vec_hsi, uv_vec_rgb, diff_AE_pixels = GeoSpatialAbstractionHSI.compute_sift_difference(hsi_image.luma_array, rgb_image.luma_array)


        image_resolution = a2
        diff_AE_meters = diff_AE_pixels*image_resolution
        return uv_vec_hsi, uv_vec_rgb, diff_AE_meters, transform_pixel_projected


    @staticmethod
    def resample_rgb_ortho_to_hsi_ortho(ref_ortho_path, hsi_composite_path, ref_ortho_reshaped_path):
        """Reproject RGB orthophoto to match the shape and projection of HSI raster.

        Parameters
        ----------
        infile : (string) path to input file to reproject
        match : (string) path to raster with desired shape and projection
        outfile : (string) path to output file tif
        """

        infile = ref_ortho_path
        match = hsi_composite_path
        outfile = ref_ortho_reshaped_path
        # open input
        with rasterio.open(infile) as src:
            src_transform = src.transform

            # open input to match
            with rasterio.open(match) as match:
                dst_crs = match.crs

                # calculate the output transform matrix
                dst_transform, dst_width, dst_height = calculate_default_transform(
                    src.crs,  # input CRS
                    dst_crs,  # output CRS
                    match.width,  # input width
                    match.height,  # input height
                    *match.bounds,  # unpacks input outer boundaries (left, bottom, right, top)
                )

            # set properties for output
            dst_kwargs = src.meta.copy()
            dst_kwargs.update({"crs": dst_crs,
                               "transform": dst_transform,
                               "width": dst_width,
                               "height": dst_height,
                               "nodata": 0})
            
            # open output
            with rasterio.open(outfile, "w", **dst_kwargs) as dst:
                # iterate through bands and write using reproject function
                for i in range(1, src.count + 1):
                    reproject(
                        source=rasterio.band(src, i),
                        destination=rasterio.band(dst, i),
                        src_transform=src.transform,
                        src_crs=src.crs,
                        dst_transform=dst_transform,
                        dst_crs=dst_crs,
                        resampling=Resampling.cubic)

    def resample_dem_to_hsi_ortho(dem_path, hsi_composite_path, dem_reshaped):
        """Reproject a file to match the shape and projection of existing raster.

        Parameters
        ----------
        infile : (string) path to input file to reproject
        match : (string) path to raster with desired shape and projection
        outfile : (string) path to output file tif
        """

        infile = dem_path
        match = hsi_composite_path
        outfile = dem_reshaped
        # open input
        with rasterio.open(infile) as src:
            src_transform = src.transform

            # open input to match
            with rasterio.open(match) as match:
                dst_crs = match.crs

                # calculate the output transform matrix
                dst_transform, dst_width, dst_height = calculate_default_transform(
                    src.crs,  # input CRS
                    dst_crs,  # output CRS
                    match.width,  # input width
                    match.height,  # input height
                    *match.bounds,  # unpacks input outer boundaries (left, bottom, right, top)
                )

            # set properties for output
            dst_kwargs = src.meta.copy()
            dst_kwargs.update({"crs": dst_crs,
                               "transform": dst_transform,
                               "width": dst_width,
                               "height": dst_height,
                               "nodata": 0})
            #print("Coregistered to shape:", dst_height, dst_width, '\n Affine', dst_transform)
            # open output
            with rasterio.open(outfile, "w", **dst_kwargs) as dst:
                # iterate through bands and write using reproject function
                for i in range(1, src.count + 1):
                    reproject(
                        source=rasterio.band(src, i),
                        destination=rasterio.band(dst, i),
                        src_transform=src.transform,
                        src_crs=src.crs,
                        dst_transform=dst_transform,
                        dst_crs=dst_crs,
                        resampling=Resampling.cubic)

    @staticmethod
    def compute_sift_difference(gray1, gray2):
        """Simply matches two grayscale images using SIFT"""

        gray1 = (gray1 - np.min(gray1)) / (np.max(gray1) - np.min(gray1))
        gray2 = (gray2 - np.min(gray2)) / (np.max(gray2) - np.min(gray2))

        gray1 = (gray1 * 255).astype(np.uint8)
        gray2 = (gray2 * 255).astype(np.uint8)


        # Find the keypoints and descriptors with SIFT
        sift = cv.SIFT_create()
        kp2, des2 = sift.detectAndCompute(gray2, None)
        kp1, des1 = sift.detectAndCompute(gray1, None)


        FLANN_INDEX_KDTREE = 1
        index_params = dict(algorithm=FLANN_INDEX_KDTREE, trees=5)
        search_params = dict(checks=50)
        flann = cv.FlannBasedMatcher(index_params, search_params)
        matches = flann.knnMatch(des1, des2, k=2)
        
        good = []
        # https://docs.opencv.org/4.x/dc/dc3/tutorial_py_matcher.html
        # store all the good matches as per Lowe's ratio test (0.7). 
        # Cranking up gives more. We changed 0.8 to 0.85 to get more matches
        for m, n in matches:
            if m.distance < 0.75 * n.distance:
                good.append(m)

        draw_params = dict(matchColor=(0, 255, 0),  # draw matches in green color
                           flags=2)

        diff_u = np.zeros(len(good))
        diff_v = np.zeros(len(good))
        uv_vec_hsi = np.zeros((len(good), 2))
        uv_vec_rgb = np.zeros((len(good), 2))
        for i in range(len(good)):
            idx2 = good[i].trainIdx
            idx1 = good[i].queryIdx
            uv1 = kp1[idx1].pt  # Slit image
            uv2 = kp2[idx2].pt  # Orthomosaic
            uv_vec_hsi[i,:] = uv1
            uv_vec_rgb[i,:] = uv2

            ## Conversion to global coordinates
            diff_u[i] = uv2[0] - uv1[0]
            diff_v[i] = uv2[1] - uv1[1]

        img3 = cv.drawMatches(gray1, kp1, gray2, kp2, good, None, **draw_params)
        """plt.imshow(img3, 'gray')
        plt.show()"""
        print(len(good))
##
        # The absolute errors
        diff_AE = np.sqrt(diff_u ** 2 + diff_v ** 2)
        
        return uv_vec_hsi, uv_vec_rgb, diff_AE



    def map_pixels_back_to_datacube(self, w_datacube):
        """The projected formats can be transformed back with four integer transforms and interpolated accordingly"""
        """As a simple strategy we perform bilinear interpolation"""

        indexes_grid = np.flip(self.indexes.reshape((self.height_rectified, self.width_rectified)), axis = 0)

        v = self.feature_uv_hsi[:, 0]
        u = self.feature_uv_hsi[:, 1]

        v_rgb = self.feature_uv_rgb[:, 0]
        u_rgb = self.feature_uv_rgb[:, 1]

        # Should transform rgb coordinates directly to world coordinates
        ## Conversion to global coordinates
        x = self.feature_uv_rgb[:, 0]
        y = self.feature_uv_rgb[:, 1]

        xoff, a, b, yoff, d, e = self.transform_pixel_projected


        xp = a * x + b * y + xoff + 0.5*a # The origin of the image coordinate system is located at 0.5,0.5
        yp = d * x + e * y + yoff + 0.5*e
        zp = np.zeros(yp.shape)
        for i in range(xp.shape[0]):
            temp = [x for x in self.raster_dem.sample([(xp[i], yp[i])])]
            zp[i] = float(temp[0])

        if self.is_global != True:
            self.features_points = np.concatenate((xp.reshape((-1,1)), yp.reshape((-1,1)), zp.reshape((-1,1))), axis = 1)
        else:
            geocsc = CRS.from_epsg(self.epsg_geocsc)
            proj = CRS.from_epsg(self.epsg_proj)
            transformer = Transformer.from_crs(proj, geocsc)
            self.features_points = np.zeros((xp.shape[0], 3))



            (x_ecef, y_ecef, z_ecef) = transformer.transform(xx=xp, yy=yp, zz=zp)

            self.features_points[:, 0] = x_ecef
            self.features_points[:, 1] = y_ecef
            self.features_points[:, 2] = z_ecef




        #

        self.v_datacube_hsi = np.zeros((v.shape[0], 4))
        self.u_datacube_hsi = np.zeros((v.shape[0], 4))

        # See wikipedia
        for i in range(4):
            if i == 0:
                u1 = np.floor(u).astype(np.int32)  # row
                v1 = np.floor(v).astype(np.int32)  # col
            elif i == 1:
                u1 = np.floor(u).astype(np.int32)  # row
                v1 = np.ceil(v).astype(np.int32)  # col
            elif i == 2:
                u1 = np.ceil(u).astype(np.int32)  # row
                v1 = np.floor(v).astype(np.int32)  # col
            else:
                u1 = np.ceil(u).astype(np.int32)  # row
                v1 = np.ceil(v).astype(np.int32)  # col

            ind_datacube_hsi = indexes_grid[u1, v1] # 1D Indexer til rå datakube

            self.v_datacube_hsi[:, i] = ind_datacube_hsi % w_datacube
            self.u_datacube_hsi[:, i]  = (ind_datacube_hsi - self.v_datacube_hsi[:, i]) / w_datacube


        self.x1_x_hsi = v - np.floor(v)
        self.y1_y_hsi = u - np.floor(u)

        #self.x1_x_rgb = v_rgb - np.floor(v_rgb)
        #self.y1_y_rgb = u_rgb - np.floor(u_rgb)


    def compute_reference_points_ecef(uv_vec_ref, transform_pixel_projected, dem_resampled_path, epsg_proj, epsg_geocsc=4978):
    
    
        """Computes ECEF reference points from a features detected on the orthomosaic and the DEM"""
        x = uv_vec_ref[:, 0]
        y = uv_vec_ref[:, 1]

        # Sample the terrain raster to get a projected position of data 
        raster_dem = rasterio.open(dem_resampled_path)
        xoff, a, b, yoff, d, e = transform_pixel_projected

        # Convert the pixel coordinates into true coordinates (e.g. UTM N/E)
        xp = a * x + b * y + xoff + 0.5*a 
        yp = d * x + e * y + yoff + 0.5*e
        zp = np.zeros(yp.shape)
        for i in range(xp.shape[0]):
            temp = [x for x in raster_dem.sample([(xp[i], yp[i])])]
            zp[i] = float(temp[0])

        # Transform points to true 3D via pyproj
        geocsc = CRS.from_epsg(epsg_geocsc)
        proj = CRS.from_epsg(epsg_proj)
        transformer = Transformer.from_crs(proj, geocsc)
        ref_points_ecef = np.zeros((xp.shape[0], 3))
        (x_ecef, y_ecef, z_ecef) = transformer.transform(xx=xp, yy=yp, zz=zp)

        ref_points_ecef[:, 0] = x_ecef
        ref_points_ecef[:, 1] = y_ecef
        ref_points_ecef[:, 2] = z_ecef
        
        return ref_points_ecef

    def compute_position_orientation_features(uv_vec_hsi, pixel_nr_image, unix_time_image, position_ecef, quaternion_ecef, time_pose, nodata):
        """Returns the positions, orientations and pixel numbers corresponding to the features. 
        Also computes a feature mask identifying features that are invalid"""
        rows = uv_vec_hsi[:, 1]
        cols = uv_vec_hsi[:, 0]

        # Determine mask
        x0 = np.floor(cols).astype(int)
        x1 = x0 + 1
        y0 = np.floor(rows).astype(int)
        y1 = y0 + 1

        # All 4 neighbors (as used by bilinear interpolation) should be valid data points
        feature_mask = np.all([pixel_nr_image[y0, x0].reshape(-1) != nodata,
               pixel_nr_image[y1, x0].reshape(-1) != nodata,
               pixel_nr_image[y0, x1].reshape(-1) != nodata,
               pixel_nr_image[y1, x1].reshape(-1) != nodata], axis=0)


        pixel_nr_vec = GeoSpatialAbstractionHSI.bilinear_interpolate(pixel_nr_image, x = cols, y = rows)
        time_vec = GeoSpatialAbstractionHSI.bilinear_interpolate(unix_time_image, x = cols, y = rows)

        pixel_vec_valid = pixel_nr_vec[feature_mask]
        time_vec_valid = time_vec[feature_mask]

        
        from scipy.interpolate import interp1d
        from gref4hsi.utils import geometry_utils as geom
        

        rotation_ecef = RotLib.from_quat(quaternion_ecef)

        # Interpolates positions linearly and quaterinons by Slerp
        position_vec, quaternion_vec = geom.interpolate_poses(time_pose, 
                                            position_ecef, 
                                            rotation_ecef,  
                                            timestamps_to=time_vec_valid)
        
        return pixel_vec_valid, time_vec_valid, position_vec, quaternion_vec, feature_mask
        

        

    @staticmethod
    def feature_matches_to_GCP(features_hsi, features_ref, pixel_nr_raster, time_raster):
        """Function to establish 2D feature positions u, j in raw cube, as well as true reference positions X, Y, Z

        :param features_hsi: _description_
        :type features_hsi: _type_
        :param features_ref: _description_
        :type features_ref: _type_
        :param pixel_nr_raster: _description_
        :type pixel_nr_raster: _type_
        :param time_raster: _description_
        :type time_raster: _type_
        """

        return None


    # COPIED from https://stackoverflow.com/questions/12729228/simple-efficient-bilinear-interpolation-of-images-in-numpy-and-python
    @staticmethod
    def bilinear_interpolate(im, x, y):
        
        x = np.asarray(x)
        y = np.asarray(y)

        x0 = np.floor(x).astype(int)
        x1 = x0 + 1
        y0 = np.floor(y).astype(int)
        y1 = y0 + 1

        x0 = np.clip(x0, 0, im.shape[1]-1);
        x1 = np.clip(x1, 0, im.shape[1]-1);
        y0 = np.clip(y0, 0, im.shape[0]-1);
        y1 = np.clip(y1, 0, im.shape[0]-1);

        Ia = im[ y0, x0 ]
        Ib = im[ y1, x0 ]
        Ic = im[ y0, x1 ]
        Id = im[ y1, x1 ]

        wa = (x1-x) * (y1-y)
        wb = (x1-x) * (y-y0)
        wc = (x-x0) * (y1-y)
        wd = (x-x0) * (y-y0)

        return wa*Ia + wb*Ib + wc*Ic + wd*Id




import configparser
import os
import sys
from collections import namedtuple

import matplotlib.pyplot as plt
import numpy as np

# Local resources:
from gref4hsi.utils.gis_tools import GeoSpatialAbstractionHSI
from gref4hsi.utils.parsing_utils import Hyperspectral



def main(iniPath):
    config = configparser.ConfigParser()
    config.read(iniPath)

    # Paths for input data
    h5_folder = config['Absolute Paths']['h5_folder']

    # Paths for output data:
    # 1) The data cube locations
    envi_cube_dir = config['Absolute Paths']['orthorectified_cube_folder']

    # 2) The RGB composite locations
    rgb_composite_dir = config['Absolute Paths']['rgb_composite_folder']

    # 3) The ancillary data locations
    anc_dir = config['Absolute Paths']['anc_folder']

    # 3) The footprints
    footprint_dir = config['Absolute Paths']['footprint_folder']

    


    # The necessary data from the H5 file for resampling the datacube and composite (fwhm is optional)
    # ECEF point cloud from georeferencing.
    h5_folder_point_cloud_ecef = config['Georeferencing']['points_ecef_crs']

    # Radiance cube
    h5_folder_radiance_cube = config['HDF.hyperspectral']['datacube']

    # Wavelength centers for all bands
    h5_folder_wavelength_centers = config['HDF.calibration']['band2wavelength']
    
    try:
        h5_folder_wavelength_widths = config['HDF.calibration']['fwhm']
    except:
        h5_folder_wavelength_widths = 'undefined'

    # The necessary data (a dictionary) from H5 file for resampling ancillary data (uses the same grid as datacube)
    anc_dict = config['Ancillary']

    # Settings having to do with coordinate reference systems, described below
    SettingsCRS = namedtuple('SettingsOrthorectification', ['epsg_geocsc', 
                                                              'epsg_proj'])
    
    config_crs = SettingsCRS(epsg_geocsc=int(config['Coordinate Reference Systems']['geocsc_epsg_export']),
                             # The epsg code of the geocentric coordinate system (ECEF)
                             epsg_proj=int(config['Coordinate Reference Systems']['proj_epsg']),
                             # The epsg code of the projected coordinate system (e.g. UTM 32 has epsg 32632 for wgs 84 ellipsoid)
                             )

    # Settings associated with orthorectification of datacube
    SettingsOrtho = namedtuple('SettingsOrthorectification', ['ground_resolution', 
                                                                            'wl_red', 
                                                                            'wl_green', 
                                                                            'wl_blue',
                                                                            'nodata_value',
                                                                            'raster_transform_method',
                                                                            'resample_rgb_only',
                                                                            'chunk_size_cube_GB',
                                                                            'wavelength_unit',
                                                                            'radiometric_unit',
                                                                            'sensor_type',
                                                                            'interleave'])
    
    config_ortho = SettingsOrtho(ground_resolution = float(config['Orthorectification']['resolutionHyperspectralMosaic']), 
                                 # Rectified grid resolution in meters
                              wl_red = float(config['General']['red_wave_length']), 
                              # Red wavelength in <wavelength_unit>
                              wl_green = float(config['General']['green_wave_length']), 
                              # Green wavelength in <wavelength_unit>
                              wl_blue = float(config['General']['blue_wave_length']), 
                              # Blue wavelength in <wavelength_unit>
                              raster_transform_method = config['Orthorectification']['raster_transform_method'], 
                              # Describes what raster transform to use. Either "north_east" or "minimal_rectangle". 
                              # Latter is fastest and most memory efficient, but support is lacking for several GIS software.
                              nodata_value = int(config['Orthorectification']['nodata']),
                              # The fill value for empty cells (select values not occcuring in cube or ancillary data)
                              resample_rgb_only = eval(config['Orthorectification']['resample_rgb_only']),
                              # Boolean being expressing whether to rectify only composite (true) or data cube and composite (false). True is fast.
                              chunk_size_cube_GB = float(config['Orthorectification']['chunk_size_cube_GB']),
                              # For large files, RAM issues could be a concern. For rectified files exeeding this size, data is written chunk-wize to a memory map.
                              wavelength_unit = config['General']['wavelength_unit'],
                              # Unit of wavelengths, often nanometer
                              radiometric_unit = config['General']['radiometric_unit'],
                              # Unit of radiance, often like <pfx1>Watts / ( [<pfx2>^2]meter^2 steradian <pfx3>meter). pfx means metric prefix.
                              # <pfx1> relates to energy and is 1, milli (10e-3) or micro (10e-6)
                              # <pfx2> relates to area and is 1 or centi (10e-2)
                              # <pfx3> relates to wavelength and is micro (10e-6) or nano (10e-9)
                              sensor_type = config['General']['sensor_type'],
                              # Brand and model
                              interleave = config['Orthorectification']['interleave']
                              # ENVI interleave: either 'bsq', 'bip' or 'bil', see:
                              # https://envi.geoscene.cn/help/Subsystems/envi/Content/ExploreImagery/ENVIImageFiles.htm
                              )


    print("\n################ Orthorectifying: ################")
    files = sorted(os.listdir(h5_folder))
    n_files= len(sorted(os.listdir(h5_folder)))
    file_count = 0
    for filename in files:
        if filename.endswith('h5') or filename.endswith('hdf'):

            progress_perc = 100*file_count/n_files
            print(f"Orthorectifying file {file_count+1}/{n_files}, progress is {progress_perc} %")
            # Path to hierarchical file
            h5_filename = h5_folder + filename

            # Read the 3D point cloud, radiance cube 
            # Extract the point cloud
            point_cloud_ecef = Hyperspectral.get_dataset(h5_filename=h5_filename,
                                                         dataset_name=h5_folder_point_cloud_ecef)
            # Need the radiance cube for resampling
            radiance_cube = Hyperspectral.get_dataset(h5_filename=h5_filename,
                                                         dataset_name=h5_folder_radiance_cube)
        
            wavelengths = Hyperspectral.get_dataset(h5_filename=h5_filename,
                                                            dataset_name=h5_folder_wavelength_centers)
            try:
                fwhm = Hyperspectral.get_dataset(h5_filename=h5_filename,
                                                            dataset_name=h5_folder_wavelength_widths)
            except KeyError:
                fwhm = np.nan


            # The code below is independent of the h5 file format. The exception is the writing of ancillary data to a form of datacube
            # Generates an object for dealing with GIS operations
            gisHSI = GeoSpatialAbstractionHSI(point_cloud=point_cloud_ecef, 
                                              transect_string=filename.split('.')[0],
                                              config_crs=config_crs)

            # The point cloud is transformed to the projected system
            gisHSI.transform_geocentric_to_projected(config_crs=config_crs)

            # Calculate the footprint of hyperspectral data
            gisHSI.footprint_to_shape_file(footprint_dir=footprint_dir)
            
            
            
            # Resample imagery (RGB composite or both)
            gisHSI.resample_datacube(radiance_cube=radiance_cube,
                                     wavelengths=wavelengths,
                                     fwhm=fwhm,
                                     envi_cube_dir=envi_cube_dir,
                                     rgb_composite_dir=rgb_composite_dir,
                                     config_ortho=config_ortho)

            

            # The ancilliary data is read from h5 files and converted into a datacube
            if eval(config['Orthorectification']['resample_ancillary']): 
                gisHSI.resample_ancillary(h5_filename=h5_filename, 
                                        anc_dict = anc_dict, 
                                        anc_dir = anc_dir,
                                        interleave=config_ortho.interleave)
        
        file_count+=1
if __name__ == '__main__':
    args = sys.argv[1:]
    iniPath = args[0]
    main(iniPath)


"""This script reads Specim data from default captures and reformats data to compatibility with pipeline"""
# Python builtins
from datetime import datetime, timedelta
import os
import glob
from pathlib import Path

# Third party libraries
import numpy as np
from pathlib import Path
from spectral import envi
import pandas as pd
import h5py
import pymap3d as pm
from scipy.interpolate import interp1d
from scipy.optimize import least_squares
from scipy.spatial.transform import Rotation as RotLib

# Library dependencies
from gref4hsi.utils.geometry_utils import CalibHSI
from gref4hsi.utils.config_utils import prepend_data_dir_to_relative_paths

ACTIVE_SENSOR_SPATIAL_PIXELS = 1024 # Constant for AFX10
ACTIVE_SENSOR_SPECTRAL_PIXELS = 448 # Constant for AFX10

class Specim():
    def __init__(self, mission_path, config):
        self.mission_path = mission_path # With slash end
        self.config = config #
    
    @staticmethod
    def fov_2_param(fov):
        """
        param fov: view angle of each pixel in degrees
        returns: A dictionary of camera parameters
        """
        # Script for reading the FOV file and converting it to a camera model
        def optimize_func(param, x_true, n_pix):

            f = param[0]  # The focal length of the lens
            k1 = param[1]  # The 1st Radial distortion
            k2 = param[2]  # The 2nd Radial distortion
            k3 = param[3]
            c_x = param[4]
            n_pix = x_true.size
            #v_c = n_pix / 2 + 1
            u = np.arange(1, n_pix + 1).reshape((-1,1))

            x_norm_lin = (u - c_x) / f
            # Distortion term if to be used
            x_norm_nonlin = -(k1 * ((u - c_x) / 1000) ** 5 + \
                            k2 * ((u - c_x) / 1000) ** 3 + \
                            k3 * ((u - c_x) / 1000) ** 2) / f

            
            x_norm = x_norm_lin + x_norm_nonlin
            diff_x = x_norm-x_true
            return diff_x.reshape(-1)
        
        
        theta_true = fov*np.pi/180

        x_true = np.array(np.tan(theta_true)).reshape((-1, 1))

        n_pix = x_true.size

        # Total angular FOV is:
        AFOV = np.abs(theta_true[0]) + np.abs(theta_true[-1])
        f_0 = (n_pix/2)/(np.tan(AFOV/2))

        param_0 = np.array([f_0, 0, 0, 0, 1 + n_pix/2])

        res = least_squares(optimize_func, param_0, args = (x_true, n_pix), x_scale = 'jac')

        param = res.x
        f = param[0]  # The focal length of the lens
        k1 = param[1]  # The 1st Radial distortion
        k2 = param[2]  # The 2nd Radial distortion
        k3 = param[3]  # The tangential distortion
        c_x = param[4] # The 

        # initialized to 0
        rotation_x = 0
        rotation_y = 0
        rotation_z = 0
        translation_x = 0
        translation_y = 0
        translation_z = 0

        param_dict = {'rx':rotation_x,
                      'ry':rotation_y,
                      'rz':rotation_z,
                      'tx':translation_x,
                      'ty':translation_y,
                      'tz':translation_z,
                      'f': f,
                      'cx': c_x,
                      'k1': k1,
                      'k2': k2,
                      'k3': k3,
                      'width': n_pix}
        return param_dict

    def read_fov_file(self, fov_file_path):
        

        

        df_fov = pd.read_csv(fov_file_path, sep = ',', header = None)

        fov_arr = df_fov.values[:, 1]

        param_dict = Specim.fov_2_param(fov = fov_arr)
        
        file_name_calib = self.config['Absolute Paths']['hsi_calib_path']

        CalibHSI(file_name_cal_xml=file_name_calib,
                 mode = 'w', 
                 param_dict = param_dict)

    def read_nav_file(self, nav_file_path, date):
        # Convert date string to datetime object
        date_obj = datetime.strptime(date, "%Y-%m-%d")
        # Open file for reading.
        with open(nav_file_path, 'r') as nav_file:
            count = 0
            imu_data = []
            gnss_data = []
            sync_data = []
            # Set to -1 to an if sentence to initialize to true
            hsi_frame_nr_prev = -1
            # Variable introduced to convert from cyclic HS frame counting to cumulative counts
            restart_counts = 0

            while True:
                count += 1

                # Get next line from file
                line = nav_file.readline()

                line_arr = line.split(sep = ',')

                if line_arr[0] == '$PASHR':
                    Timestamp_str = line_arr[1]
                    sec = float(Timestamp_str[4:])
                    minut = int(Timestamp_str[2:4])
                    hour = int(Timestamp_str[0:2])

                    # Create a timedelta object for the time
                    time_delta = timedelta(hours=hour, minutes=minut, seconds=sec)

                    # Calculate the total timedelta
                    total_timedelta = date_obj + time_delta

                    # Convert to UNIX timestamp (decimal UNIX time)
                    unix_timestamp = (total_timedelta - datetime(1970, 1, 1)) / timedelta(seconds=1)

                    TimestampAbs = unix_timestamp

                    TimestampClock = date + '_' + Timestamp_str

                    Yaw = float(line_arr[2]) # Heading wrt true north
                    Roll = float(line_arr[4])
                    Pitch = float(line_arr[5])# According to https://docs.novatel.com/OEM7/Content/SPAN_Logs/PASHR.htm


                    sigmaRoll = float(line_arr[7])
                    sigmaPitch = float(line_arr[8])  # According to https://docs.novatel.com/OEM7/Content/SPAN_Logs/PASHR.htm
                    sigmaYaw = float(line_arr[9])

                    #dlogr_imu.append_data(np.array([TimestampClock, TimestampAbs, Roll, Pitch, Yaw, sigmaRoll, sigmaPitch, sigmaYaw]))

                    imu_data_row = {'TimestampClock':TimestampClock,
                                     'TimestampAbs': TimestampAbs,
                                     'Roll': Roll,
                                     'Pitch': Pitch,
                                     'Yaw': Yaw,
                                     'sigmaRoll': sigmaRoll,
                                     'sigmaPitch': sigmaPitch,
                                     'sigmaYaw': sigmaYaw}
                    imu_data.append(imu_data_row)

                elif line_arr[0] == '$GNGGA' or line_arr[0] == '$GPGGA':
                    Timestamp_str = line_arr[1]
                    sec = float(Timestamp_str[4:])
                    minut = float(Timestamp_str[2:4])
                    hour = float(Timestamp_str[0:2])

                    # Create a timedelta object for the time
                    time_delta = timedelta(hours=hour, minutes=minut, seconds=sec)

                    # Calculate the total timedelta
                    total_timedelta = date_obj + time_delta

                    # Convert to UNIX timestamp (decimal UNIX time)
                    unix_timestamp = (total_timedelta - datetime(1970, 1, 1)) / timedelta(seconds=1)

                    TimestampAbs = unix_timestamp

                    TimestampClock = date + '_' + Timestamp_str

                    LatDDMM_MM = float(line_arr[2]) # DDMM.mm
                    LatInt = int(LatDDMM_MM/100)
                    Lat = LatInt + (LatDDMM_MM - LatInt*100)/60


                    LonDDDMM_MM = float(line_arr[4]) # DDDMM.mm
                    LonInt = int(LonDDDMM_MM / 100)
                    Lon = LonInt + (LonDDDMM_MM - LonInt * 100) / 60

                    AltMSL = float(line_arr[9])
                    AltGeoid = float(line_arr[11])  # According to https://docs.novatel.com/OEM7/Content/Logs/GPGGA.htm

                    #dlogr_gnss.append_data(np.array([TimestampClock, TimestampAbs, Lon, Lat, AltMSL, AltGeoid]))
                    gnss_data_row = {'TimestampClock':TimestampClock,
                                     'TimestampAbs': TimestampAbs,
                                     'Lon': Lon,
                                     'Lat': Lat,
                                     'AltMSL': AltMSL,
                                     'AltGeoid': AltGeoid}
                    gnss_data.append(gnss_data_row)
                elif line_arr[0] == '$SPTSMP':
                    """Recordings of time sync messages, tied to the hyperspectral frames"""
                    hsi_frame_number_new = int(line_arr[2]) - 1 # Seems they start with a 1-base (like MATLAB)

                    if hsi_frame_number_new < hsi_frame_nr_prev: # When frame counter resets
                        restart_counts += (hsi_frame_nr_prev - hsi_frame_number_new) + self.metadata_obj.fps # 9950 + 50


                    hsi_frame_number = hsi_frame_number_new
                    hsi_frame_nr_prev = hsi_frame_number_new
                    prev_time = TimestampAbs

                    # In the file, fram counts go 1, 51, 101,... 9951, 1 ... in a cyclic manner. So to get cumulative counts we add the restart_counts
                    sync_data_row = {'HsiFrameNum': hsi_frame_number + restart_counts,
                                     'TimestampAbs': prev_time,
                                     'TimestampClock': TimestampClock}
                    sync_data.append(sync_data_row)




                # if line is empty
                # end of file is reached
                if not line:
                    break
        
        self.imu_data = imu_data
        self.gnss_data = gnss_data
        self.sync_data = sync_data

"""Writer for the h5 file format using a dictionary. The user provides h5 hierarchy paths as values and keys are the names given to the attributes of the specim object.
A similar write process could be applied to metadata."""
def specim_object_2_h5_file(h5_filename, h5_tree_dict, specim_object):
    with h5py.File(h5_filename, 'w', libver='latest') as f:
        for attribute_name, h5_hierarchy_item_path in h5_tree_dict.items():
            #print(attribute_name)
            dset = f.create_dataset(name=h5_hierarchy_item_path, 
                                            data = getattr(specim_object, attribute_name))            

def add_byte_order_to_envi_header(header_file_path, byte_order_value):
    """Function added to remedy lacking byte-order entry in header files of radiometric calibration data

    :param header_file_path: _description_
    :type header_file_path: _type_
    :param byte_order_value: _description_
    :type byte_order_value: _type_
    """
    # Read the existing ENVI header
    with open(header_file_path, 'r') as f:
        header_lines = f.readlines()

    # Look for the line where you want to add "byte order"
    for i, line in enumerate(header_lines):
        if line.startswith('byte order'):
            # If it already exists, update the value
            header_lines[i] = f'byte order = {byte_order_value}\n'
            break
    else:
        # If it doesn't exist, add it to the end of the header
        header_lines.append(f'byte order = {byte_order_value}\n')

    # Save the updated header
    with open(header_file_path, 'w') as f:
        f.writelines(header_lines)



def main(config, config_specim):
    
    # Function that calibrates data and reforms to h5 file format.
    dtype = config_specim.dtype_datacube
    transect_chunk_size = config_specim.lines_per_chunk
    cal_dir = config_specim.cal_dir
   
    mission_dir = config_specim.specim_raw_mission_dir

    mission_name = Path(mission_dir).name
    out_dir = config_specim.reformatted_missions_dir
    config_file_path = os.path.join(out_dir, config_specim.config_file_name)

    specim_object = Specim(mission_path=mission_dir, config=config)

    # Patterns for searching data cube files
    PATTERN_ENVI = '*.hdr'

    # Expects the captured data to reside in capture subfolder
    capture_dir = os.path.join(mission_dir, 'capture')

    # Path for searching (avoid explicit )
    search_path_envi = os.path.normpath(os.path.join(capture_dir, PATTERN_ENVI))
    envi_hdr_file_path = glob.glob(search_path_envi)[0]

    # Read out data
    spectral_image_obj = envi.open(envi_hdr_file_path)

    """import spectral as sp
    rgb_im = spectral_image_obj[:,:,]
    sp.imshow(spectral_image_obj, (73, 50, 24))"""

    # Read all meta data from header file (currently hard coded, but could be avoided I guess)
    class Metadata:
        pass

    metadata_obj = Metadata()
    # Could do this in iteration, but it would need to specify type (I think)
    metadata_obj.autodarkstartline = int(spectral_image_obj.metadata['autodarkstartline'])
    metadata_obj.n_lines = int(spectral_image_obj.metadata['lines'])
    metadata_obj.n_bands = int(spectral_image_obj.metadata['bands'])
    metadata_obj.n_pix = int(spectral_image_obj.metadata['samples'])
    metadata_obj.t_exp_ms = float(spectral_image_obj.metadata['tint'])
    metadata_obj.fps = float(spectral_image_obj.metadata['fps'])
    metadata_obj.description = spectral_image_obj.metadata['description']
    metadata_obj.file_type = spectral_image_obj.metadata['file type']
    metadata_obj.sensor_type = spectral_image_obj.metadata['sensor type']
    metadata_obj.acquisition_date = spectral_image_obj.metadata['acquisition date']
    metadata_obj.sensorid = spectral_image_obj.metadata['sensorid']
    metadata_obj.interleave = spectral_image_obj.metadata['interleave']
    metadata_obj.data_type = spectral_image_obj.metadata['data type']
    
    # Derived from knowledge of the sensor (sensor constants that could differ for other sensors)
    metadata_obj.binning_spatial = int(ACTIVE_SENSOR_SPATIAL_PIXELS/metadata_obj.n_pix)
    metadata_obj.binning_spectral = int(ACTIVE_SENSOR_SPECTRAL_PIXELS/metadata_obj.n_bands)

    # Allow Specim Methods to access metadata
    specim_object.metadata_obj = metadata_obj 
    # -

    # Based on the binning info, we can locate relevant calibration files, including 1) spectral, 2) geometric, 3) radiometric, and dark frame (from capture).
    #
    # We start with the spectral calibration:

    # Comprehensive band info (center, fwhm) is found in "cal_dir/wlcal<spectral binning>b_fwhm.wls"
    # Pattern follows structure
    PATTERN_BAND_INFO = '*'+ str(metadata_obj.binning_spectral) + 'b_fwhm.wls'
    # Linux-CLI search for file.
    search_path_bands = os.path.normpath(os.path.join(cal_dir, PATTERN_BAND_INFO))
    band_file_path = glob.glob(search_path_bands)[0]

    # Read the file as csv (although it has no columns)
    df_bands = pd.read_csv(band_file_path, header=None, sep = '\s+')
    df_bands.columns = ['Wavelength_nm', 'FWHM_nm']

    # Extract band data
    specim_object.wavelengths = np.array(df_bands['Wavelength_nm'])
    specim_object.fwhm = np.array(df_bands['FWHM_nm'])

    # -

    # We will here go through the loading of Specim geometric camera calibration. 
    # The first step is to load the angular Field-of-View file (AFOV) from the manufacturer. 
    # Then boresight angles and lever arms can be set, if relevant.

    # +
    # Pixel-directions is found in "cal_dir/FOV_****_<spatial binning>b.txt" 
    PATTERN_FOV = 'FOV*' + '_' +  str(metadata_obj.binning_spatial) + 'b.txt'

    # Search for fov file.
    search_path_fov = os.path.normpath(os.path.join(cal_dir, PATTERN_FOV))
    fov_file_path = glob.glob(search_path_fov)[0]

    # Calculate a camera model (zero boresight) based on FOV file.
    # Writes parameters to an *xml file specified by config

    # Read fov data (one angle for each pixel)
    df_fov = pd.read_csv(fov_file_path, header=None, sep = ',')

    df_fov.columns = ['pixel_nr', 'view_angle_deg', 'Unknown']

    specim_object.view_angles = np.array(df_fov['view_angle_deg'])
    
    # Simply converts view angles to camera parameters for a 1D line camera model
    param_dict = Specim.fov_2_param(fov = specim_object.view_angles)
    
    # param_dict are the full camera model parameters, but defaults rotation matrix and translation

    # User set rotation matrix
    R_hsi_body = config_specim.rotation_matrix_hsi_to_body

    r_zyx = RotLib.from_matrix(R_hsi_body).as_euler('ZYX', degrees=False)

    # Euler angle representation (any other would do too)
    param_dict['rz'] = r_zyx[0]
    param_dict['ry'] = r_zyx[1]
    param_dict['rx'] = r_zyx[2]

    # Vector from origin of HSI to body origin, expressed in body
    # User set
    t_hsi_body = config_specim.translation_body_to_hsi
    param_dict['tz'] = t_hsi_body[0]
    param_dict['ty'] = t_hsi_body[1]
    param_dict['tz'] = t_hsi_body[2]

    camera_calib_xml_dir = config['Absolute Paths']['calib_folder']

    file_name_xml = 'HSI_' + str(metadata_obj.binning_spatial) + 'b.xml'
    xml_cal_write_path = camera_calib_xml_dir + file_name_xml

    CalibHSI(file_name_cal_xml= xml_cal_write_path, 
                    mode = 'w', 
                    param_dict = param_dict)


    # Set value in config file:
    config.set('Relative Paths', 'hsi_calib_path', value = xml_cal_write_path)

    

    # Write the config object 
    with open(config_file_path, 'w') as configfile:
            config.write(configfile)
    # -

    


    # The next step is to read in radiometric frame.

    # +
    """Extract radiometric frame from dedicated file"""

    PATTERN_ENVI_CAL = '*_' +str(metadata_obj.binning_spectral) + 'x' +  str(metadata_obj.binning_spatial) + '.hdr'

    search_path_envi_cal = os.path.normpath(os.path.join(cal_dir, PATTERN_ENVI_CAL))
    ENVI_CAL_HDR_FILE_PATH = glob.glob(search_path_envi_cal)[0]

    # For allowing spectral module to read
    RAD_CAL_BYTE_ORDER = 0

    add_byte_order_to_envi_header(header_file_path=ENVI_CAL_HDR_FILE_PATH, byte_order_value=RAD_CAL_BYTE_ORDER)

    ENVI_CAL_IMAGE_FILE_PATH = ENVI_CAL_HDR_FILE_PATH.split('.')[0] + '.cal' # SPECTRAL does not expect this suffix by default

    # For some reason, the byte order is needed
    radiometric_image_obj = envi.open(ENVI_CAL_HDR_FILE_PATH, image = ENVI_CAL_IMAGE_FILE_PATH)

    cal_n_lines = int(radiometric_image_obj.metadata['lines'])
    cal_n_bands = int(radiometric_image_obj.metadata['bands'])
    cal_n_pix = int(radiometric_image_obj.metadata['samples'])

    # For calibration
    radiometric_frame = radiometric_image_obj[:,:,:].reshape((cal_n_pix, cal_n_bands))

    specim_object.radiometric_frame = radiometric_frame
    # -

    # The next step is to load the darkframes

    # +
    """1) Crop the hyperspectral data according to the start-stop lines. 2) Write datacube to appropriate directory"""
    # To ensure that the plots actually do appear in this notebook:
    # %matplotlib qt

    # Establish dark frame data (at end of recording)
    data_dark = spectral_image_obj[metadata_obj.autodarkstartline:metadata_obj.n_lines, :, :]
    dark_frame = np.median(data_dark, axis = 0)

    specim_object.dark_frame = dark_frame

    # -

    # The navigation data is given as messages is a *.nav file. Locate the file and parse it into a suitable format.

    # +
    # Extract the starting/stopping lines
    START_STOP_DIR = os.path.join(mission_dir, 'start_stop_lines')
    
    # Allow software to function if start_stop_lines not specified
    if not os.path.exists(START_STOP_DIR):
        os.mkdir(START_STOP_DIR)
        # Chunk according to chunk size
        start_lines = np.arange(start=0,
                                stop=metadata_obj.autodarkstartline, 
                                step=transect_chunk_size)

        stop_lines = np.arange(start=transect_chunk_size, 
                               stop=metadata_obj.autodarkstartline + transect_chunk_size, 
                               step=transect_chunk_size)
        
        stop_lines[-1] = metadata_obj.autodarkstartline # No need to georeference dark data

        # Make dataframe
        start_stop_data = {'line_start' : start_lines,
                           'line_stop' : stop_lines}
        
        df_start_stop = pd.DataFrame(start_stop_data)

        df_start_stop.to_csv(path_or_buf = os.path.join(START_STOP_DIR, 'start_stop_lines.txt'), sep=' ')



    PATTERN_START_STOP = '*.txt'
    

    search_path_lines_start_stop = os.path.normpath(os.path.join(START_STOP_DIR, PATTERN_START_STOP))
    LINES_START_STOP_FILE_PATH = glob.glob(search_path_lines_start_stop)[0]

    df_start_stop = pd.read_csv(filepath_or_buffer=LINES_START_STOP_FILE_PATH, header=0, sep=' ')



    # +
    # Now read the *.nav file
    NAV_PATTERN = '*.nav'

    search_path_nav = os.path.normpath(os.path.join(capture_dir, NAV_PATTERN))
    nav_file_path = glob.glob(search_path_nav)[0]

    date_string = metadata_obj.acquisition_date.split(': ')[1] # yyyy-mm-dd
    # Parse the position/orientation messages
    specim_object.read_nav_file(nav_file_path=nav_file_path, date = date_string)



    # -

    # Calculate the frame timestamps from sync data

    # +
    


    df_imu = pd.DataFrame(specim_object.imu_data)
    df_gnss = pd.DataFrame(specim_object.gnss_data)
    df_sync_hsi = pd.DataFrame(specim_object.sync_data)
    
    # Define the time stamps of HSI frames (special fix for Specim)
    sync_frames = df_sync_hsi['HsiFrameNum']
    sync_times = df_sync_hsi['TimestampAbs']
    hsi_frames = np.arange(metadata_obj.autodarkstartline)
    hsi_timestamps_total = interp1d(x = sync_frames, y = sync_times, fill_value = 'extrapolate')(x = hsi_frames)


    # For ease, let us interpolate position data to imu time (allows one timestamp for nav data)
    imu_time = df_imu['TimestampAbs']

    # Drop the specified regular clock time (as it is not needed)
    df_gnss = df_gnss.drop(columns=['TimestampClock'])

    # Interpolate each column in GNSS data based on 'imu_time'
    interpolated_values = {
        column: np.interp(imu_time, df_gnss['TimestampAbs'], df_gnss[column])
        for column in df_gnss.columns if column != 'TimestampAbs'
    }

    # Create a new DataFrame with the interpolated values
    df_gnss_interpolated = pd.DataFrame({'time': imu_time, **interpolated_values})

    # The position defined in geodetic coordinates
    lat = np.array(df_gnss_interpolated['Lat']).reshape((-1,1))
    lon = np.array(df_gnss_interpolated['Lon']).reshape((-1,1))
    ellipsoid_height = np.array(df_gnss_interpolated['AltMSL'] + df_gnss_interpolated['AltGeoid']).reshape((-1,1))

    # Assumes WGS-84 (default GNSS frame)
    x, y, z = pm.geodetic2ecef(lat = lat, lon = lon, alt = ellipsoid_height, deg=True)

    # Lastly, calculate the roll, pitch, yaw
    roll = np.array(df_imu['Roll']).reshape((-1,1))
    pitch = np.array(df_imu['Pitch']).reshape((-1,1))
    yaw = np.array(df_imu['Yaw']).reshape((-1,1))

    # Roll pitch yaw are stacked in attribute eul_zyx. The euler angles with rotation order ZYX are Yaw Pitch Roll
    specim_object.eul_zyx = np.concatenate((roll, pitch, yaw), axis = 1)

    # Position is stored as ECEF cartesian coordinates (mutually orthogonal axis) instead of spherioid-like lon, lat, alt
    specim_object.position_ecef = np.concatenate((x,y,z), axis = 1)
    specim_object.nav_timestamp = imu_time
    specim_object.t_exp_ms = metadata_obj.t_exp_ms

    # -

    # Last preprocessing step is writing to h5 files. 
    

    h5_dict_write = {'eul_zyx' : config['HDF.raw_nav']['eul_zyx'],
            'position_ecef' : config['HDF.raw_nav']['position'],
            'nav_timestamp' : config['HDF.raw_nav']['timestamp'],
            'radiance_cube': config['HDF.hyperspectral']['datacube'],
            't_exp_ms': config['HDF.hyperspectral']['exposuretime'],
            'hsi_timestamps': config['HDF.hyperspectral']['timestamp'],
            'view_angles': config['HDF.calibration']['view_angles'],
            'wavelengths' : config['HDF.calibration']['band2wavelength'],
            'fwhm' : config['HDF.calibration']['fwhm'],
            'dark_frame' : config['HDF.calibration']['darkframe'],
            'radiometric_frame' : config['HDF.calibration']['radiometricframe']}

    # Time to write all the data to a h5 file



    # # Chunking the recording to user defined sizes and writing it to disk

    # +
    # Define h5 file name
    h5_folder = config['Absolute Paths']['h5_folder']

    # Each line in start_stop defines a transect
    n_transects = df_start_stop.shape[0]
    for transect_number in range(n_transects):
        start_line = df_start_stop['line_start'][transect_number]
        stop_line = df_start_stop['line_stop'][transect_number]

        n_chunks = int(np.ceil((stop_line-start_line)/transect_chunk_size))

        
        for chunk_number in range(n_chunks):
            chunk_start_idx = start_line + transect_chunk_size*chunk_number

            if chunk_number == n_chunks-1:
                chunk_stop_idx = stop_line
            else:
                chunk_stop_idx = chunk_start_idx + transect_chunk_size

            #print(f'Chunk from line {chunk_start_idx} to line {chunk_stop_idx} is being written')
            data_cube = spectral_image_obj[chunk_start_idx:chunk_stop_idx, :, :]
            # Calibration equation
            specim_object.radiance_cube = ( (data_cube - dark_frame)*radiometric_frame/(metadata_obj.t_exp_ms/1000) ).astype(dtype = dtype) # 4 Byte
            specim_object.hsi_timestamps = hsi_timestamps_total[chunk_start_idx:chunk_stop_idx]

            # Possible to name files with <PREFIX>_<time_start>_<Transect#>_<Chunk#>.h5
            h5_filename = h5_folder + mission_name + '_transectnr_' + str(int(transect_number)) + '_chunknr_' + str(int(chunk_number)) + '.h5'

            specim_object_2_h5_file(h5_filename=h5_filename, h5_tree_dict=h5_dict_write, specim_object=specim_object)












import pathlib
import setuptools

setuptools.setup(
    name='gref4hsi',
    version='0.2.3',    
    description='A Python package for for georeferencing and orthorectifying hyperspectral imagery',
    long_description=pathlib.Path("README.md").read_text(),
    long_description_content_type = "text/markdown",
    url='https://github.com/havardlovas/gref4hsi',
    author='Haavard Snefjellaa Loevaas',
    author_email='havard.s.lovas@ntnu.no',
    license='EUPL-1.2',
    install_requires=[
        "opencv-python",
        "dill",
        "geopandas",
        "h5py",
        "matplotlib",
        "numpy",
        "open3d",
        "pandas",
        "Pillow",
        "pymap3d",
        "pyproj",
        "pyvista",
        "pyvistaqt",
        "scikit-learn",
        "scipy",
        "shapely",
        "spectral",
        "xmltodict",
        "pykdtree",
        "trimesh",
        "rtree",
        "embreex",
        "ephem"
    ],

    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: European Union Public Licence 1.2 (EUPL 1.2)',  
        'Operating System :: Microsoft :: Windows :: Windows 10',        
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
    ],
    python_requires='>3.7',
    packages=setuptools.find_packages(),
    include_package_data=True
)



# Standard python library
import configparser
import sys
import os
import argparse
from collections import namedtuple
import pathlib
import yaml

if os.name == 'nt':
    # Windows OS
    base_fp = 'D:'
    home = 'C:/Users/haavasl'
elif os.name == 'posix':
    # This Unix-like systems inl. Mac and fLinux
    base_fp = '/media/haavasl/Expansion'
    home = '/home/haavasl'

# Use this if working with the github repo to do quick changes to the module
module_path = os.path.join(home, 'VsCodeProjects/gref4hsi/')
if module_path not in sys.path:
    sys.path.append(module_path)

# Local resources
from gref4hsi.scripts import georeference
from gref4hsi.scripts import orthorectification
from gref4hsi.scripts import coregistration
from gref4hsi.utils import parsing_utils, specim_parsing_utils
from gref4hsi.utils import visualize
from gref4hsi.utils.config_utils import prepend_data_dir_to_relative_paths, customize_config


import numpy as np

"""
This script is meant to be used for testing the processing pipeline of airborne HI data from the Specim AFX10 instrument.
"""

# Make it simple to swap when working a bit on windows and a bit on Linux



def main(config_yaml, specim_mission_folder, geoid_path, config_template_path, lab_calibration_path, fast_mode = False):
    # Read flight-specific yaml file
    with open(config_yaml, 'r') as file:  
        config_data = yaml.safe_load(file)
    
    
    # assigning the arguments to variables for simple backwards compatibility
    SPECIM_MISSION_FOLDER = specim_mission_folder
    EPSG_CODE = config_data['mission_epsg']
    RESOLUTION_ORTHOMOSAIC = config_data['resolution_orthomosaic']
    CALIBRATION_DIRECTORY = lab_calibration_path
    
    
    dem_fold = os.path.join(specim_mission_folder, "dem")

    if not os.path.exists(dem_fold):
        print('DEM folder does not exist so Geoid is used as terrain instead')
        TERRAIN_TYPE = "geoid"
    else:
        if not os.listdir(dem_fold):
            #print(f"The folder '{dem_fold}' is empty so Geoid is used as terrain instead.")
            TERRAIN_TYPE = "geoid"
        else:
            # If there is a folder and it is not empty
            # Find the only file that is there
            files = [f for f in os.listdir(dem_fold) if f not in ('.', '..')]
            DEM_PATH = os.path.join(dem_fold, files[0])
            #print(f"The file '{DEM_PATH}' is used as terrain.")
            TERRAIN_TYPE = "dem_file"
    
    
    # Do coregistration if there is an orthomosaic to compare under "orthomosaic"
    #do_coreg = True
    ortho_ref_fold = os.path.join(specim_mission_folder, "orthomosaic")

    if not os.path.exists(ortho_ref_fold):
        print('Coregistration is not done, as there was no reference orthomosaic')
        
    else:
        if not os.listdir(ortho_ref_fold):
            print('Coregistration is not done, as there was no reference orthomosaic')
        else:
            # If there is a folder and it is not empty
            # Find the only file that is there
            ortho_ref_file = [f for f in os.listdir(ortho_ref_fold) if f not in ('.', '..')][0]
            do_coreg = True
            print(f"The file '{ortho_ref_file}' is used as as reference orthomosaic.")
            
    
    
    GEOID_PATH = geoid_path

    # Settings associated with preprocessing of data from Specim Proprietary data to pipeline-compatible data
    SettingsPreprocess = namedtuple('SettingsPreprocessing', ['dtype_datacube', 
                                                                            'lines_per_chunk', 
                                                                            'specim_raw_mission_dir',
                                                                            'cal_dir',
                                                                            'reformatted_missions_dir',
                                                                            'rotation_matrix_hsi_to_body',
                                                                            'translation_body_to_hsi',
                                                                            'config_file_name'])

    config_specim_preprocess = SettingsPreprocess(dtype_datacube = np.float32, # The data type for the datacube
                                lines_per_chunk= 8000,  # Raw datacube is chunked into this many lines. GB_per_chunk = lines_per_chunk*n_pixels*n_bands*4 bytes
                                specim_raw_mission_dir = SPECIM_MISSION_FOLDER, # Folder containing several mission
                                cal_dir = CALIBRATION_DIRECTORY,  # Calibration directory holding all calibrations at all binning levels
                                reformatted_missions_dir = os.path.join(SPECIM_MISSION_FOLDER, 'processed'), # The fill value for empty cells (select values not occcuring in cube or ancillary data)
                                rotation_matrix_hsi_to_body = np.array([[0, 1, 0],
                                                                        [-1, 0, 0],
                                                                        [0, 0, 1]]), # Rotation matrix R rotating so that vec_body = R*vec_hsi.
                                translation_body_to_hsi = np.array([0, 0, 0]), # Translation t so that vec_body_to_object = vec_hsi_to_object + t
                                # For large files, RAM issues could be a concern. For rectified files exeeding this size, data is written chunk-wize to a memory map.
                                config_file_name = 'configuration.ini')



    # Where to place the config
    DATA_DIR = config_specim_preprocess.reformatted_missions_dir
    config_file_mission = os.path.join(DATA_DIR, 'configuration.ini')


    # Set the data directory for the mission, and create empty folder structure
    prepend_data_dir_to_relative_paths(config_path=config_template_path, DATA_DIR=DATA_DIR)

    # Non-default settings
    custom_config = {'General':
                        {'mission_dir': DATA_DIR,
                        'model_export_type': TERRAIN_TYPE, # Ray trace onto geoid
                        'max_ray_length': 150}, # Max distance in meters from spectral imager to seafloor. Specim does not fly higher

                    'Coordinate Reference Systems':
                        {'proj_epsg' : EPSG_CODE, # The projected CRS UTM 32, common on mainland norway
                        'geocsc_epsg_export' : 4978, # 3D cartesian system for earth consistent with GPS frame (but inconsistent with eurasian techtonic plate)
                        'dem_epsg' : EPSG_CODE, # (Optional) If you have a DEM this can be used
                        'pos_epsg_orig' : 4978}, # The CRS of the positioning data we deliver to the georeferencing

                    'Orthorectification':
                        {'resample_rgb_only': True, # True can be good choice for speed during DEV
                         'resample_ancillary': True,
                        'resolutionhyperspectralmosaic': RESOLUTION_ORTHOMOSAIC, # Resolution in m
                        'raster_transform_method': 'north_east'}, # North-east oriented rasters.
                    
                    'HDF.raw_nav': {
                        'rotation_reference_type' : 'eul_ZYX', # The vehicle orientations are given in Yaw, Pitch, Roll from the NAV system
                        'is_global_rot' : False, # The vehicles orientations from NAV system are Yaw, Pitch, Roll
                        'eul_is_degrees' : True}, # And given in degrees
                    'Absolute Paths': {
                        'geoid_path' : GEOID_PATH,
                        #'geoid_path' : 'data/world/geoids/egm08_25.gtx',
                        'dem_path' : DEM_PATH,
                        'orthomosaic_reference_folder' : os.path.join(specim_mission_folder, "orthomosaic"),
                        'ref_ortho_reshaped' : os.path.join(DATA_DIR, "Intermediate", "RefOrthoResampled"),
                        'ref_gcp_path' : os.path.join(DATA_DIR, "Intermediate", "gcp.csv"),
                        'calib_file_coreg' : os.path.join(DATA_DIR, "Output", "HSI_coreg.xml"),
                        # (above) The georeferencing allows processing using norwegian geoid NN2000 and worldwide EGM2008. Also, use of seafloor terrain models are supported. '
                        # At the moment refractive ray tracing is not implemented, but it could be relatively easy by first ray tracing with geoid+tide, 
                        # and then ray tracing from water
                        #'tide_path' : 'D:/HyperspectralDataAll/HI/2022-08-31-060000-Remoy-Specim/Input/tidevann_nn2000_NMA.txt'
                        },
                    
                    # If coregistration is done, then the data must be stored after processing somewhere
                    'HDF.coregistration': {
                            'position_ecef': 'processed/coreg/position_ecef',
                            'quaternion_ecef' : 'processed/coreg/quaternion_ecef'
                        },
                    # These are the ancillary layers to be orthorectified (can select from all entities in "Georeferencing")
                    'Ancillary': {
                            'pixel_nr_grid': 'processed/georef/pixel_nr_grid', # h5 path
                            'unix_time_grid' : 'processed/georef/unix_time_grid' # h5 path
                        }
                    
    }

    if TERRAIN_TYPE == 'geoid':
        custom_config['Absolute Paths']['geoid_path'] = GEOID_PATH
        #'geoid_path' : 'data/world/geoids/egm08_25.gtx'
    elif TERRAIN_TYPE == 'dem_file':
        custom_config['Absolute Paths']['dem_path'] = DEM_PATH

    
    # No need to orthorectify the data cube initially when coregistration with RGB composites is done
    if do_coreg:
        custom_config['Orthorectification']['resample_rgb_only'] = True
    
    # 
    if fast_mode:
        custom_config['Orthorectification']['resample_rgb_only'] = True
        custom_config['Orthorectification']['resolutionhyperspectralmosaic'] = 1


    # Customizes the config file according to settings
    customize_config(config_path=config_file_mission, dict_custom=custom_config)


    config = configparser.ConfigParser()
    config.read(config_file_mission)

    # This function parses raw specim data including (spectral, radiometric, geometric) calibrations and nav data
    # into an h5 file. The nav data is written to "raw/nav/" subfolders, whereas hyperspectral data and calibration data 
    # written to "processed/hyperspectral/" and "processed/calibration/" subfolders
    """specim_parsing_utils.main(config=config,
                              config_specim=config_specim_preprocess)"""
    
    # Time interpolates and reformats the pose (of the vehicle body) to "processed/nav/" folder.
    parsing_utils.export_pose(config_file_mission)
    
    # Formats model to triangular mesh and an earth centered earth fixed / geocentric coordinate system
    parsing_utils.export_model(config_file_mission)

    # Commenting out the georeference step is fine if it has been done

    
    ## Visualize the data 3D photo model from RGB images and the time-resolved positions/orientations
    #visualize.show_mesh_camera(config, show_mesh = True, show_pose = True, ref_frame='ENU')

    # Step 1: Direct georeferencing
    """georeference.main(config_file_mission)

    # Step 2: Orthorectify the direct georeferenced data (incl metadata)
    orthorectification.main(config_file_mission)"""
    


    # Step 3 (optional): Coregistration
    # The coregistration compares to the reference (requires a reference orthomosaic) and optimizes geometric parameters which are used to re-georeference.
    if do_coreg:
        # Coregistration requires that Step 1 and 2 were performed and that resample anc = True
        # Match RGB composite to reference, find features and following data, ground control point (gcp) list, for each feature pair:
        # reference point 3D (from reference), position/orientation of vehicle (using resampled time) and pixel coordinate (using resampled pixel coordinate)
        """coregistration.main(config_file_mission, mode='compare', is_calibrated = False)

        # The gcp list allows reprojecting reference points and evaluate the reprojection error,
        # which is used to optimize static geometric parameters (e.g. boresight, camera model...) or dynamic geometric parameters (time-varying nav errors).
        # Settings are currently in coregistration script
        coregistration.main(config_file_mission, mode='calibrate', is_calibrated = False)

        # Only resample RGB and aux data 
        custom_config['Orthorectification']['resample_rgb_only'] = True
        customize_config(config_path=config_file_mission, dict_custom=custom_config)

        # Re-georeference with coregistred parameters
        georeference.main(config_file_mission, use_coreg_param=True)
        
        # Coregister this stuff
        orthorectification.main(config_file_mission)

        # Second round of comparison
        coregistration.main(config_file_mission, mode='compare', is_calibrated = True)"""
        
        # Check bulk
        coregistration.main(config_file_mission, mode='calibrate', is_calibrated = True)



if __name__ == "__main__":
    
    
    # Globally accessible files:
    geoid_path = os.path.join(home, "VsCodeProjects/gref4hsi/data/world/geoids/egm08_25.gtx")
    config_template_path = os.path.join(home, "VsCodeProjects/gref4hsi/data/config_examples/configuration_specim.ini")
    lab_calibration_path = os.path.join(base_fp, "Specim/Lab_Calibrations")
    
    # Third flight: The configuration file and recording folder on drive
    """specim_mission_folder = os.path.join(base_fp, r"Specim/Missions/2022-08-31-Remoy/remoy_202208311435_ntnu_hyperspectral_74m")
    config_yaml = os.path.join(specim_mission_folder, "config.seabee.yaml")

    # Run 
    main(str(config_yaml), str(specim_mission_folder), geoid_path, config_template_path, lab_calibration_path)"""

   # Second flight
    """specim_mission_folder = os.path.join(base_fp, r"Specim/Missions/2022-08-31-Remoy/remoy_202208311040_ntnu_hyperspectral_74m")
    config_yaml = os.path.join(specim_mission_folder, "config.seabee.yaml")

    # Run 
    main(str(config_yaml), str(specim_mission_folder), geoid_path, config_template_path, lab_calibration_path)"""

    # First flight
    specim_mission_folder = os.path.join(base_fp, r"Specim/Missions/2022-08-31-Remoy/remoy_202208310800_ntnu_hyperspectral_74m")
    config_yaml = os.path.join(specim_mission_folder, "config.seabee.yaml")

    # Run 
    main(str(config_yaml), str(specim_mission_folder), geoid_path, config_template_path, lab_calibration_path)


# Built-ins
import configparser
import json
import os
import sys

# Third party
from scipy.spatial.transform import Rotation as RotLib
import numpy as np
import pyvista as pv
import h5py

# Lib resources:
from gref4hsi.utils.geometry_utils import CameraGeometry, CalibHSI
from gref4hsi.utils.parsing_utils import Hyperspectral
from gref4hsi.utils import visualize


def cal_file_to_rays(filename_cal):
        # See paper by Sun, Bo, et al. "Calibration of line-scan cameras for precision measurement." Applied optics 55.25 (2016): 6836-6843.
        # Loads line camera parameters for the hyperspectral imager from an xml file.

        # Certain imagers deliver geometry "per pixel". This can be resolved by fitting model parameters.
        calHSI = CalibHSI(file_name_cal_xml=filename_cal)
        f = calHSI.f
        cx = calHSI.cx

        # Radial distortion parameters
        k1 = calHSI.k1
        k2 = calHSI.k2

        # Tangential distortion parameters
        k3 = calHSI.k3

        # Translation (lever arm) of HSI with respect to vehicle frame
        trans_x = calHSI.tx
        trans_y = calHSI.ty
        trans_z = calHSI.tz

        # Rotation of HSI (boresight) with respect to vehicle navigation frame (often defined by IMU or RGB cam)
        rot_x = calHSI.rx
        rot_y = calHSI.ry
        rot_z = calHSI.rz

        # Number of pixels
        n_pix = calHSI.w

        # Define camera model array.
        u = np.arange(0, n_pix) + 1

        # Express uhi ray directions in uhi frame using line-camera model
        x_norm_lin = (u - cx) / f

        x_norm_nonlin = -(k1 * ((u - cx) / 1000) ** 5 + \
                          k2 * ((u - cx) / 1000) ** 3 + \
                          k3 * ((u - cx) / 1000) ** 2) / f

        x_norm = x_norm_lin + x_norm_nonlin

        p_dir = np.zeros((len(x_norm), 3))

        # Rays are defined in the HI frame with positive z down
        p_dir[:, 0] = x_norm
        p_dir[:, 2] = 1

        rot_hsi_ref_eul = np.array([rot_z, rot_y, rot_x])

        rot_hsi_ref_obj = RotLib.from_euler(seq = 'ZYX', angles = rot_hsi_ref_eul, degrees=False)

        translation_ref_hsi = np.array([trans_x, trans_y, trans_z])

        intrinsic_geometry_dict = {'translation_ref_hsi': translation_ref_hsi,
                                   'rot_hsi_ref_obj': rot_hsi_ref_obj,
                                   'ray_directions_local': p_dir}
        
        # Notably, one could compress the information by expressing the ray directions in the body frame

        return intrinsic_geometry_dict

def define_hsi_ray_geometry(pos_ref_ecef, quat_ref_ecef, time_pose, intrinsic_geometry_dict):
        """Instantiate a camera geometry object from the h5 pose data"""


        pos = pos_ref_ecef # Reference positions in ECEF
        rot_obj = RotLib.from_quat(quat_ref_ecef) # Reference orientations wrt ECEF
        
        ray_directions_local = intrinsic_geometry_dict['ray_directions_local']
        translation_ref_hsi = intrinsic_geometry_dict['translation_ref_hsi']
        rot_hsi_ref_obj = intrinsic_geometry_dict['rot_hsi_ref_obj']

        hsi_geometry = CameraGeometry(pos=pos, rot=rot_obj, time=time_pose, is_interpolated=True)
        
        
        hsi_geometry.intrinsicTransformHSI(translation_ref_hsi=translation_ref_hsi, rot_hsi_ref_obj = rot_hsi_ref_obj)

        hsi_geometry.defineRayDirections(dir_local = ray_directions_local)

        return hsi_geometry

def write_intersection_geometry_2_h5_file(hsi_geometry, config, h5_filename):
    # Write all intersection data (ancilliary) that could be relevant
    
    dict_ancilliary = config['Georeferencing']
    # Dictionary keys correspond to CameraGeometry attribute names (e.g. hsi_geometry.key), while values correspond to h5 data set paths.
    
    with h5py.File(h5_filename, 'a', libver='latest') as f:
        for attribute_name, h5_hierarchy_item_path in dict_ancilliary.items():
            if attribute_name != 'folder':
                if h5_hierarchy_item_path in f:
                    del f[h5_hierarchy_item_path]
                dset = f.create_dataset(name=h5_hierarchy_item_path, 
                                                data = getattr(hsi_geometry, attribute_name))


# Function called to apply standard processing on a folder of files
def main(iniPath, viz = False, use_coreg_param = False):
    config = configparser.ConfigParser()
    config.read(iniPath)

    # Paths to 3D mesh ply file 
    path_mesh = config['Absolute Paths']['model_path']

    # Directory of H5 files
    dir_r = config['Absolute Paths']['h5_folder']

    

    

    # Timestamps here
    h5_folder_time_pose = config['HDF.processed_nav']['timestamp']

    # Use the regular parameters from nav system and manufacturer:
    # The path to the XML file
    hsi_cal_xml = config['Absolute Paths']['hsi_calib_path']

    # Position is stored here in the H5 file
    h5_folder_position_ecef = config['HDF.processed_nav']['position_ecef']

    # Quaternion is stored here in the H5 file
    h5_folder_quaternion_ecef = config['HDF.processed_nav']['quaternion_ecef']

    if use_coreg_param:
        print('Using coregistred parameters for georeferencing')

        # Set the camera model to the calibrated one if it exists
        hsi_cal_xml_coreg = config['Absolute Paths']['calib_file_coreg']
        if os.path.exists(hsi_cal_xml_coreg):
             hsi_cal_xml = hsi_cal_xml_coreg

        # Optimized position
        h5_folder_position_ecef_coreg = config['HDF.coregistration']['position_ecef']

        # Quaternion 
        h5_folder_quaternion_ecef_coreg = config['HDF.coregistration']['quaternion_ecef']
        

    
    
    
    # The path to the Tide file (if necessary and available)
    try:
        path_tide = config['Absolute Paths']['tide_path']
    except Exception as e:
        path_tide = 'Undefined'
    
    # Maximal allowed ray length
    max_ray_length = float(config['General']['max_ray_length'])

    mesh = pv.read(path_mesh)

    model_meta_path = path_mesh.split('.')[0] + '_meta.json' 
    with open(model_meta_path, "r") as f:
        # Load the JSON data from the file
        metadata_mesh = json.load(f)
        mesh_off_x = metadata_mesh['offset_x']
        mesh_off_y = metadata_mesh['offset_y']
        mesh_off_z = metadata_mesh['offset_z']
        # Mesh is translated by this much
        mesh_trans = np.array([mesh_off_x, mesh_off_y, mesh_off_z]).astype(np.float64)

    
    print("\n################ Georeferencing: ################")
    files = sorted(os.listdir(dir_r))
    n_files= len(sorted(os.listdir(dir_r)))
    file_count = 0
    for filename in files:
        if filename.endswith('h5') or filename.endswith('hdf'):

            progress_perc = 100*file_count/n_files
            print(f"Georeferencing file {file_count+1}/{n_files}, progress is {progress_perc} %")

            # Path to hierarchical file
            h5_filename = dir_r + filename

            # Read h5 file
            hyp = Hyperspectral(h5_filename, config)

            if use_coreg_param:
                 try:
                    # Use the coregistred dataset if it exists
                    print('Using coregistred position')
                    pos_ref_ecef = Hyperspectral.get_dataset(h5_filename=h5_filename,
                                                    dataset_name= h5_folder_position_ecef_coreg)
                 except:
                    # If not use the original
                    pos_ref_ecef = Hyperspectral.get_dataset(h5_filename=h5_filename,
                                                    dataset_name= h5_folder_position_ecef)
                 try:
                    # Use the coregistred quaternion-dataset if it exists
                    print('Using coregistred quaternion')
                    quat_ref_ecef = Hyperspectral.get_dataset(h5_filename=h5_filename,
                                                    dataset_name= h5_folder_quaternion_ecef_coreg)
                 except:
                     # If not use the original
                    quat_ref_ecef = Hyperspectral.get_dataset(h5_filename=h5_filename,
                                                    dataset_name= h5_folder_quaternion_ecef)
            else:

                # Just use the regular navigation data
                # If not use the original
                pos_ref_ecef = Hyperspectral.get_dataset(h5_filename=h5_filename,
                                                dataset_name= h5_folder_position_ecef)
                # Extract the ecef orientations for each frame
                quat_ref_ecef = Hyperspectral.get_dataset(h5_filename=h5_filename,
                                                                dataset_name=h5_folder_quaternion_ecef)
            # Extract the timestamps for each frame
            time_pose = Hyperspectral.get_dataset(h5_filename=h5_filename,
                                                            dataset_name= h5_folder_time_pose)


            # Using the cal file, we can define lever arm, boresight and camera model geometry (in dictionary)
            intrinsic_geometry_dict = cal_file_to_rays(filename_cal=hsi_cal_xml)

            
            # Define the rays in ECEF for each frame. 
            hsi_geometry = define_hsi_ray_geometry(pos_ref_ecef, 
                                    quat_ref_ecef, 
                                    time_pose,
                                    intrinsic_geometry_dict = intrinsic_geometry_dict)

            
            hsi_geometry.intersect_with_mesh(mesh = mesh, max_ray_length=max_ray_length, mesh_trans = mesh_trans)

            print(f'Mesh translation is {mesh_trans}')
            
            # Computes the view angles in the local NED. Computationally intensive as local NED is defined for each intersection
            hsi_geometry.compute_view_directions_local_tangent_plane()

            # Computes the sun angles in the local NED. Computationally intensive as local NED is defined for each intersection
            hsi_geometry.compute_sun_angles_local_tangent_plane()

            hsi_geometry.compute_tide_level(path_tide, tide_format = 'NMA')

            hsi_geometry.compute_elevation_mean_sealevel(source_epsg = config['Coordinate Reference Systems']['geocsc_epsg_export'], 
                                                         geoid_path = config['Absolute Paths']['geoid_path'])
            
            write_intersection_geometry_2_h5_file(hsi_geometry=hsi_geometry, config = config, h5_filename=h5_filename)

            hsi_geometry.write_rgb_point_cloud(config = config, hyp = hyp, transect_string = filename.split('.')[0])

            if viz:
                 visualize.show_projected_hsi_points(HSICameraGeometry=hsi_geometry, 
                                                     config=config, 
                                                     transect_string = filename.split('.')[0])

            


            
            

            #from scripts import visualize
            #visualize.show_projected_hsi_points(HSICameraGeometry=hsi_geometry, config=config, transect_string = filename.split('.')[0])

        file_count+=1
            












if __name__ == '__main__':
    args = sys.argv[1:]
    iniPath = args[0]
    main(iniPath)







from collections import namedtuple

import numpy as np
import pandas as pd
from scipy.interpolate import interp1d
from scipy.special import expi
from scipy.special import erf
from scipy.special import erfi
from scipy.optimize import fsolve
import matplotlib.pyplot as plt

def fresnel(theta_i, n_i, n_t, is_transmittance, method):
    # This is used for computing the transmittance at the interfaces
    # First Snells:
    if method == 1:
        theta_t = np.arcsin((n_i / n_t)*np.sin(theta_i))

        c_i = np.cos(theta_i)
        c_t = np.cos(theta_t)

        R_s = np.abs((n_i*c_i - n_t*c_t)/(n_i*c_i + n_t*c_t))**2

        R_p = np.abs((n_i*c_t - n_t*c_i)/(n_i*c_t + n_t*c_i))**2

        # Then snells:
        R_F = 0.5*(R_s + R_p)


    elif method == 2:
        theta_t = np.arcsin((n_i / n_t) * np.sin(theta_i))

        r1 = (np.sin(theta_i - theta_t)/np.sin(theta_i + theta_t))**2
        r2 = (np.tan(theta_i - theta_t) / np.tan(theta_i + theta_t)) ** 2

        R_F = 0.5 * (r1 + r2)
    if is_transmittance:
        return 1 - R_F
    else:
        return R_F




def solid_angle_ratio(theta_w, n_w):
    """Computes the in-air angle corresponding to an in-water angle using Snell's law"""
    n_a = 1 # Approximate n_a = 1
    theta_a = np.arcsin((n_w / n_a) * np.sin(theta_w)) # Snells law
    return theta_a

def immersion_factor(theta_w, n_w, n_g, pixel_nr):
    """Calculates the immersion factor for imager with planar glass port in underwater hyperspectral imaging"""
    n_a = 1

    theta_a = np.arcsin((n_w / n_a) * np.sin(theta_w))

    t_ag = fresnel(theta_i=theta_a, n_i = n_a, n_t = n_g, is_transmittance=1, method=1)

    t_wg = fresnel(theta_i=theta_w, n_i=n_w, n_t=n_g, is_transmittance=1, method=1)

    I_f_LU = ((n_w**2)/(n_a**2) ) * (t_ag/t_wg)

    I_f = I_f_LU[pixel_nr.reshape(-1), :]

    return I_f



def radiance_conversion(RF, t_exp, pixel_nr):

    rad_2_DN_LU = RF*t_exp

    rad_2_DN = rad_2_DN_LU[pixel_nr.reshape(-1)] # Evaluate for relevant pixels
    return rad_2_DN



def beam_pattern(dir_s0, dir, sigma, I_0, I_hat):
    n = dir.shape[0]
    k = I_hat.shape[0]
    # A gaussian beam pattern
    dot_prod = np.dot(dir_s0, np.transpose(dir))

    len_vec = np.linalg.norm(dir_s0)*np.linalg.norm(dir, axis = 1)

    theta_si = np.arccos(dot_prod/len_vec)



    z = (theta_si)/sigma

    Intensity = I_0*np.exp(-0.5*z**2)

    Intensity_Spectral = np.multiply(Intensity.reshape((-1, 1)), I_hat.reshape((1, -1)))

    return Intensity/I_0, Intensity_Spectral

def compute_gamma(dir, normals):
    """Compute the angle between a normal vector and an incoming ray"""
    dot_prod = np.einsum('ij, ij -> i', normals, -dir)
    len_vec = np.linalg.norm(dir, axis = 1) * np.linalg.norm(normals, axis = 1)
    gamma = np.arccos(dot_prod/len_vec)
    return gamma.reshape((-1, 1))

def compute_n_g(wlen):
    n_g = 1.4424 + (7.1661/(wlen - 144.170))
    return n_g

def compute_n_w(wlen, method):
    if method == 1:
        n_w = 1.325147 + (6.6096/(wlen - 137.1924))
    elif method == 2:
        T = 10
        a = -0.000001501562500
        b = 0.000000107084865
        c = -0.000042759374989
        d = -0.000160475520686
        e = 1.398067112092424
        # Function of Slocum

        n_w = a * T ** 2 + b * wlen ** 2 + c * T + d * wlen + e

    return n_w

def compute_scattering_integral(k, x):

    Ei = expi(x = -2*x*k)

    integral = -(2*k*x*Ei + np.exp(-k*x)) / x

    return integral

def compute_backscatter(B_p, b, b_w, b_p, d_pix, pos_seabed, p_si, dir_s0, sigma_l, I_0, I_hat, k, pixel_nr):
    # Step 1 create a lookup table for pixelnumber and perp distance
    z_res = 0.05

    z_max = pos_seabed[:, 2].max()

    # Partition
    z = (np.arange(0, np.ceil(z_max/z_res) + 1)*z_res).reshape((-1, 1, 1))

    r_si_b = z * d_pix - p_si # t by m by 3

    l_si_b = np.linalg.norm(r_si_b, axis =2).reshape((r_si_b.shape[0], r_si_b.shape[1], 1)) # t by m by 1 or something

    r_b_h = -z*d_pix # t by m by 3

    l_b_h = np.linalg.norm(r_b_h, axis=2).reshape((r_si_b.shape[0], r_si_b.shape[1], 1))  # t by m by 1 or something

    BP_norm, BP = beam_pattern(dir_s0=dir_s0, dir=r_si_b.reshape((-1, 3)), sigma=sigma_l, I_0=I_0, I_hat=I_hat).reshape((r_si_b.shape[0], r_si_b.shape[1], I_hat.shape[0]))

    E_i = BP*(1/(l_si_b**2))*np.exp(-k*l_si_b) # Values of incident irradiance for all directions and all perpendicular


    # distances t by m by k. Part of the integrand

    # Compute psi for all t by m cells. Check that angles are on the right side of 90 deg

    dot_prod = np.einsum('jk, ijk -> ij', d_pix, r_si_b)
    len_vec = np.linalg.norm(d_pix, axis=1) * np.linalg.norm(r_si_b, axis=2)
    psi = np.pi - np.arccos(dot_prod / len_vec)

    beta_p_interp = fournier_forand(B_p)  # Assumed constant backscatter fraction across wl. Is a log-log interpolator

    beta_p = np.exp( beta_p_interp(np.log(psi)) ).reshape((r_si_b.shape[0], r_si_b.shape[1], 1)) # Due to the interpolation in log-log domain. t by m

    beta_w = compute_beta_w(psi).reshape((r_si_b.shape[0], r_si_b.shape[1], 1)) # t by m

    beta_tot = (1/b)*(b_p*beta_p + b_w*beta_w) # t by m by k

    integrand = E_i*beta_tot*np.exp(-k*l_b_h) # t by m by k

    integral = np.cumsum(integrand, axis = 0)*z_res # Summing along the z axis weighted by z_res. Should return a t by m by k.



    L_b_LU = b*integral # Should return a t by m by k.

    # Can compute row of a given measurement by:
    rows = np.round(pos_seabed[:, 2]/z_res).astype(np.int64).reshape(-1)
    cols = pixel_nr.reshape(-1)

    L_b = L_b_LU[rows, cols, :] # Returns the appropriate spectra n by k

    #print(stop - start)

    return L_b









def compute_light_source_integral(sigma):
    """Computes the relation between spectral radiant flux P [W] and peak intensity I_0 [W/sr] for a directionally gaussian beam
    with sigma standard decviation"""
    constant = np.exp(-0.5 * sigma ** 2) * sigma

    term1 = -0.626657j * erf(0 + 0.707107j * sigma)
    term2 = 0.626657j * erf(1.11072 / sigma + 0.707107j * sigma)
    term3 = -0.626657 * erfi((0.707107 * (sigma ** 2 + 1.5708j)) / sigma)
    term4 = 0.626657 * erfi(0.707107 * sigma + 0j)

    return np.real(constant * (term1 + term2 + term3 + term4))*2*np.pi

def fournier_forand(B_p):
    """Calculates the phase function values for a given backscatter fraction"""

    # Partition data into equal 100 bins from 10 degrees to 100:
    logPsi = np.linspace(np.log(10*np.pi/180), np.log(180*np.pi/180), 100)
    psi = np.exp(logPsi)
    # Convert B_p to FF parameters
    mu, n = compute_FF_params(B_p)
    #
    nu = (3 - mu) / 2

    delta = (4 * (np.sin(psi / 2)) ** 2) / (3 * (n - 1) ** 2) # Correct, Mobley 2002 eq 2

    delta_180 = 4 / (3 * (n - 1) ** 2)

    beta_ff_t1 = 1 / (4 * np.pi * ((1 - delta) ** 2) * (delta ** nu))  # Coeff 1, correct

    beta_ff_t2 = nu * (1 - delta) - (1 - delta ** nu) # Correct

    beta_ff_t3 = delta * (1 - delta ** nu) - nu * (1 - delta) # *Correct

    beta_ff_t4 = 1 / (np.sin(psi / 2) ** 2) # Correct

    beta_ff_t5 = (1 - delta_180 ** nu) / (16 * np.pi * (delta_180 - 1) * delta_180 ** nu)

    beta_ff_t6 = 3 * np.cos(psi) ** 2 - 1

    beta_ff = beta_ff_t1 * (beta_ff_t2 + beta_ff_t3 * beta_ff_t4) \
              + beta_ff_t5 * beta_ff_t6  ## Formula t1*(t2+t3*t4) + t5*t6

    # Instantiate an interpolation object that describes the function
    beta_ff_interp = interp1d(x = logPsi, y = np.log(beta_ff), kind = 'cubic', fill_value='extrapolate')

    return beta_ff_interp

def compute_FF_params(B_p):
    mu_0 = 4 # Midpoint between max and min allowed values
    mu = fsolve(eq_backscatter_param, mu_0, args = (B_p,))
    n = 1.01 + 0.1542*(mu-3) # Eq 621, p 202 ocean optics
    return mu, n

    #

def eq_backscatter_param(mu, B_p_true):
    n = 1.01 + 0.1542*(mu-3) # Eq 621, p 202 ocean optics
    nu = (3-mu)/2 # eq 2 mobley 2002

    psi_90 = np.pi/2

    delta_90 = (4/( 3*(n-1)**2 ))*np.sin(psi_90/2)**2  # eq 2 mobley 2002

    num = (1 - delta_90**(nu+1) - 0.5*(1 - delta_90**nu))
    den = (1 - delta_90)*delta_90**nu

    B_p = 1 - num/den

    return B_p_true - B_p



def compute_beta_w(psi):
    """The scattering phase function of pure water"""
    # Psi is an array of scattering angles
    # Analytical Formula from Ocean Optics p. 180
    return 0.0608*(1 + 0.925*(np.cos(psi))**2)




class Radiometry():
    def __init__(self, config, data1, data2):
        # Read in necessary information. i.e. water absorption and water scattering coefs
        path_iop_water = 'C:/Users/haavasl/PycharmProjects/newGit/TautraReflectanceTools/Misc/iop_pure_water.txt'
        df_light = pd.read_csv('C:/Users/haavasl/PycharmProjects/newGit/TautraReflectanceTools/Misc/DataMultiSeaLite.txt',
                         header=None)

        #df_light = pd.read_csv(
        #    'C:/Users/haavasl/PycharmProjects/newGit/TautraReflectanceTools/Misc/Multi_SeaLite_spectrum.csv',
        #    header=0)

        df_iop = pd.read_csv(path_iop_water, header=9, sep=',') # Should be interpolated

        df_coral = pd.read_csv(
            'C:/Users\haavasl\PycharmProjects/newGit\TautraReflectanceTools\Misc\Coral_spectra_lab.csv', sep=';')

        df_refl_plaque = pd.read_csv('C:/Users\haavasl\PycharmProjects/newGit\TautraReflectanceTools\Misc/reflectance_plaque.csv')

        interpol_coral = interp1d(df_coral.iloc[:, 0].astype(np.float64),
                                np.transpose(df_coral.iloc[:, 2:4].astype(np.float64)), fill_value='extrapolate')
        interpol_iop = interp1d(df_iop.iloc[:, 0].astype(np.float64), np.transpose(df_iop.iloc[:, 1:3].astype(np.float64)))
        interpol_light = interp1d(df_light.iloc[:, 0], np.transpose(df_light.iloc[:, 1]), fill_value='extrapolate')

        interpol_refl = interp1d(df_refl_plaque.iloc[:, 0], np.transpose(df_refl_plaque.iloc[:, 1]))

        refl_gt = interpol_refl(data1.bands)

        I_hat = interpol_light(data1.bands) # Total spectral radiant flux
        a_w, b_w = interpol_iop(data1.bands)
        s1, s2 = interpol_coral(data1.bands)
        refl_gt_coral = 0.5*(s1 + s2) # Mean of measurements

        n_w = compute_n_w(wlen = data1.bands, method = 2)
        n_g = compute_n_g(wlen = data1.bands)

        # data contains RadData(rad = rad_2, points = points_local_2, normals=normals_local_2, pixel_nr=pixel_2, bands=bands_2, angles)

        Var = namedtuple('Var', ['points1', 'normals1', 'pixel_nr1', 'bands1', 'angles1','RF1','t_exp1', 'points2', 'normals2', 'pixel_nr2', 'bands2', 'angles2', 'RF2', 't_exp2'])

        self.Var =  Var(points1 = data1.points, normals1=data1.normals, pixel_nr1=data1.pixel_nr, bands1=data1.bands, angles1=data1.angles, RF1 = data1.RF, t_exp1 = data1.t_exp,
                         points2 = data2.points, normals2=data2.normals, pixel_nr2=data2.pixel_nr, bands2=data2.bands, angles2=data2.angles, RF2 = data2.RF, t_exp2 = data2.t_exp)


        # Structure the data in 1) constants, 2) measurements y, 3) variables x and 4) parameters beta
        Meas = namedtuple('Meas', ['rad1', 'rad2'])
        self.Meas = Meas(rad1 = data1.rad, rad2=data2.rad)



        Const = namedtuple('Const', ['S', 'a_w', 'b_w', 'n_g', 'n_w', 'dir_s1', 'dir_s2', 'pos_s1', 'pos_s2', 'I_hat', 'refl_gt', 'refl_gt_coral'])
        self.Const = Const(S = 0.015, a_w = a_w, b_w = b_w, n_g = n_g, n_w=n_w, dir_s1=np.array([0,0,1]), dir_s2=np.array([0,0,1]),
                           pos_s1 = np.array([0.35, 0, 0]), pos_s2 = np.array([-0.35, 0, 0]), I_hat = I_hat, refl_gt= refl_gt, refl_gt_coral = refl_gt_coral)


        Param = namedtuple('Param', ['G', 'X', 'Y', 'B_p', 'alpha', 'I_0', 'sigma_l', 'k_l'])
        self.Param = Param(G = 0.1, X = 0.1, Y = 1, B_p=0.01, alpha = 0, I_0 = 1, sigma_l = 1, k_l = 0)
    def run_forward_model(self, is_skogn, model = 'forward'):
        #print('This function is yet to be made')
        if model == 'two way':
            self.refl_1 = self.two_way_model_simple(id = 1, is_skogn = is_skogn)
            self.refl_2 = self.two_way_model_simple(id = 2, is_skogn = is_skogn)
        if model == 'forward':
            self.diff_dc1 = self.forward_model_simple(id=1, is_skogn=is_skogn)
            self.diff_dc2 = self.forward_model_simple(id=2, is_skogn=is_skogn)



            # Could also propagate the uncertainty into the model by perturbing


    def set_parameters(self, G, X, Y, B_p, alpha, I_0, sigma_l, k_l):
        #self.Param = Param(G=0.1, X=0.1, Y=1, alpha=0, I_0=1, sigma_l=1)
        Param = namedtuple('Param', ['G', 'X', 'Y', 'B_p', 'alpha', 'I_0', 'sigma_l', 'k_l'])
        self.Param = Param(G=G, X=X, Y=Y, B_p = B_p, alpha=alpha, I_0=I_0, sigma_l=sigma_l, k_l = k_l)





    def two_way_model_simple(self, id, is_skogn):
        # The formward model computes
        # Parameters
        G = self.Param.G
        X = self.Param.X
        Y = self.Param.Y
        B_p = self.Param.B_p


        I_0 = self.Param.I_0
        sigma_l = self.Param.sigma_l
        alpha = self.Param.alpha
        #B_inf1 = self.Param.B_inf
        #B_inf2 = self.Param.B_inf
        #p1  = self.Param.p1
        #p2 = self.Param.p2
        #theta_s1 = self.Param.theta_s1


        if id == 1:
            # Measurements
            L_air = self.Meas.rad1
            # Variables
            wlen = self.Var.bands1
            pos_seabed = self.Var.points1
            norm_vec = self.Var.normals1
            theta_w_dot = self.Var.angles1
            pixel_nr = self.Var.pixel_nr1
        elif id == 2:
            # Measurements
            L_air = self.Meas.rad2
            # Variables
            wlen = self.Var.bands2
            pos_seabed = self.Var.points2
            norm_vec = self.Var.normals2
            theta_w_dot = self.Var.angles2
            pixel_nr = self.Var.pixel_nr2


        # Constants
        a_ph = 0
        a_w = self.Const.a_w
        b_w = self.Const.b_w
        S = self.Const.S
        dir_s10 = self.Const.dir_s1
        dir_s20 = self.Const.dir_s2
        pos_s1 = self.Const.pos_s1
        pos_s2 = self.Const.pos_s2
        n_w = self.Const.n_w.reshape((1, -1))
        n_g = self.Const.n_g.reshape((1, -1))


        I_hat_non_scaled = self.Const.I_hat


        k_scale = self.Param.k_l

        I_hat = I_hat_non_scaled * (1 +  k_scale*(wlen - wlen[100]))





        # Eq. 9a)-9f)
        b_bw = 0.5 * b_w #Water is isotrophic
        a_g = G*np.exp(-S*(wlen - 440))
        b_bp = X * (400 / wlen) ** Y
        b_p = b_bp/B_p
        a = a_w + a_ph + a_g
        b = b_w + b_p

        b_b = b_bw + b_bp

        k = (a + b_b).reshape((1, -1))

        import matplotlib.pyplot as plt

        #plt.show()
#
        #plt.plot(wlen, I_hat*I_0)
        #plt.show()
        # Vector from source to bottom
        dir_s1 = pos_seabed - pos_s1
        dir_s2 = pos_seabed - pos_s2
        # Length of those
        R_s1 = np.linalg.norm(dir_s1, axis = 1).reshape((-1, 1))
        R_s2 = np.linalg.norm(dir_s2, axis = 1).reshape((-1, 1))
        # Length of camera-bottom vector
        R_c = np.linalg.norm(pos_seabed, axis = 1).reshape((-1, 1))

        # Returns an n by k vector where k are the spectral
        BP1_norm, BP1 = beam_pattern(dir_s0=dir_s10, dir = dir_s1, sigma= sigma_l, I_0 = I_0, I_hat = I_hat)
        BP2_norm, BP2 = beam_pattern(dir_s0=dir_s20, dir = dir_s2, sigma=sigma_l, I_0=I_0, I_hat = I_hat)

        # Returns an n by 1 vector of cosines
        gamma_s1 = compute_gamma(dir = dir_s1, normals = norm_vec)
        gamma_s2 = compute_gamma(dir=dir_s2, normals = norm_vec)

        if id == 1:
            self.cos_gamma_s1_t1 = np.cos(gamma_s1)
            self.cos_gamma_s2_t1 = np.cos(gamma_s2)
            self.BP1_norm_t1 = BP1_norm
            self.BP2_norm_t1 = BP2_norm
            self.R_s1_t1 = R_s1
            self.R_s2_t1 = R_s2

        elif id == 2:
            # Measurements
            self.cos_gamma_s1_t2 = np.cos(gamma_s1)
            self.cos_gamma_s2_t2 = np.cos(gamma_s2)
            self.BP1_norm_t2 = BP1_norm
            self.BP2_norm_t2 = BP2_norm
            self.R_s1_t2 = R_s1
            self.R_s2_t2 = R_s2

        # Simple computation of incident light
        E_i1 = BP1 * np.cos(gamma_s1) * np.exp(-k * R_s1) / (R_s1 ** 2)
        E_i2 = BP2 * np.cos(gamma_s2) * np.exp(-k * R_s2) / (R_s2 ** 2)




        import matplotlib.pyplot as plt


        if is_skogn == False:
            E_i = E_i1 + E_i2
        else:
            E_i = E_i1 # Only this light source that is illuminating the scene

        # Calculate from measurement to seafloor
        theta_w = theta_w_dot - alpha # Intersection angle with respect to glass

        I_f = immersion_factor(theta_w=theta_w, n_w = n_w, n_g = n_g, pixel_nr = pixel_nr)

        L_w = np.einsum('ij,ij -> ij', L_air, I_f)

        #if id == 2:
        #    from scipy.ndimage import gaussian_filter1d
        #    plt.hist(L_air.reshape(-1), 100)
        #    plt.show()

        # Compute backscatter
        include_backscatter = False
        if include_backscatter:
            x_pix = np.tan(theta_w_dot).reshape((-1, 1))
            y_pix = np.zeros(x_pix.shape)
            z_pix = np.ones(x_pix.shape)
            d_pix = np.concatenate((x_pix, y_pix, z_pix), axis = 1)
            k_bands = k.shape[1]

            if is_skogn:
                L_s1 = compute_backscatter(B_p=B_p, b=b.reshape((1, 1, k_bands)), b_w=b_w.reshape((1, 1, k_bands)),
                                           b_p=b_p.reshape((1, 1, k_bands)), d_pix=d_pix, pos_seabed=pos_seabed,
                                           p_si=pos_s1, dir_s0=dir_s10, sigma_l=sigma_l,
                                           I_0=I_0, I_hat=I_hat, k=k.reshape((1, 1, k_bands)), pixel_nr=pixel_nr)
                L_s = L_s1# The total scattered light
            else:

                L_s1 = compute_backscatter(B_p = B_p, b = b.reshape((1, 1, k_bands)), b_w = b_w.reshape((1, 1, k_bands)),
                                           b_p = b_p.reshape((1, 1, k_bands)), d_pix = d_pix, pos_seabed = pos_seabed,
                                           p_si = pos_s1, dir_s0 = dir_s10, sigma_l = sigma_l,
                                           I_0 = I_0, I_hat = I_hat, k= k.reshape((1, 1, k_bands)), pixel_nr = pixel_nr)

                L_s2 = compute_backscatter(B_p=B_p, b=b.reshape((1, 1, k_bands)), b_w=b_w.reshape((1, 1, k_bands)),
                                           b_p=b_p.reshape((1, 1, k_bands)), d_pix=d_pix, pos_seabed=pos_seabed,
                                           p_si=pos_s2, dir_s0=dir_s20, sigma_l=sigma_l,
                                           I_0=I_0, I_hat=I_hat, k=k.reshape((1, 1, k_bands)), pixel_nr=pixel_nr)
                L_s = L_s1 + L_s2  # The total scattered light



        else:
            L_s1 = 0
            L_s2 = 0
            L_s = L_s1 + L_s2

        #poly_dir_1 = p1 * (theta_w_dot - theta_s1) ** 2 + p2
        #poly_dir_2 = p1 * (theta_w_dot + theta_s1) ** 2 + p2

        #L_s1 = I_0 * int_1 * b_b * poly_dir_1
        #L_s2 = I_0 * int_2 * b_b * poly_dir_2



        L_d = L_w - L_s # The direct light

        #import matplotlib.pyplot as plt

        L_r = L_d * np.exp(k * R_c) # Backwards attenuation
        #plt.hist(R_c, 50)
        #plt.show()
        rho = np.pi * L_r / E_i

        #print(np.sum(np.isnan(rho)))
        #print(np.sum(np.isinf(rho)))
        return rho

    def forward_model_simple(self, id, is_skogn):
        """This forward model propagates light for a set of Param. The error is evaluated in digital counts"""
        ## IOP params
        G = self.Param.G
        X = self.Param.X
        Y = self.Param.Y
        # B

        ## Light source params
        I_0 = self.Param.I_0
        sigma_l = self.Param.sigma_l

        ## Immersion factor params
        alpha = self.Param.alpha


        if id == 1:
            # Measurements
            L_air = self.Meas.rad1

            # Variables
            wlen = self.Var.bands1
            pos_seabed = self.Var.points1
            norm_vec = self.Var.normals1
            theta_w_dot = self.Var.angles1
            pixel_nr = self.Var.pixel_nr1
            RF = self.Var.RF1
            t_exp = self.Var.t_exp1
        elif id == 2:
            # Measurements
            L_air = self.Meas.rad2
            # Variables
            wlen = self.Var.bands2
            pos_seabed = self.Var.points2
            norm_vec = self.Var.normals2
            theta_w_dot = self.Var.angles2
            pixel_nr = self.Var.pixel_nr2
            RF = self.Var.RF2
            t_exp = self.Var.t_exp2


        # Constants
        a_ph = 0
        a_w = self.Const.a_w
        b_bw = self.Const.b_w
        S = self.Const.S
        dir_s10 = self.Const.dir_s1
        dir_s20 = self.Const.dir_s2
        pos_s1 = self.Const.pos_s1
        pos_s2 = self.Const.pos_s2

        n_w = self.Const.n_w.reshape((1, -1))
        n_g = self.Const.n_g.reshape((1, -1))
        #P_0 = self.Const.P_0 # Defines total Spectral flux in air
        I_hat = self.Const.I_hat
        rho_gt = self.Const.refl_gt.reshape((1, -1))

        # IOP definitions from Lee
        a_g = G*np.exp(-S*(wlen - 440))

        a = a_w + a_ph + a_g

        b_bp_dot = X*(400/wlen)**Y

        b_b = b_bw + b_bp_dot # eq 18

        k = (a + b_b).reshape((1, -1))


        # Vector from source to bottom
        dir_s1 = pos_seabed - pos_s1
        dir_s2 = pos_seabed - pos_s2
        # Length of those
        R_s1 = np.linalg.norm(dir_s1, axis = 1).reshape((-1, 1))
        R_s2 = np.linalg.norm(dir_s2, axis = 1).reshape((-1, 1))


        import matplotlib.pyplot as plt


        # Length of camera-bottom vector
        R_c = np.linalg.norm(pos_seabed, axis = 1).reshape((-1, 1))

        # Returns an n by k vector where k are the spectral
        BP1_norm, BP1 = beam_pattern(dir_s0=dir_s10, dir = dir_s1, sigma= sigma_l, I_0 = I_0, I_hat = I_hat)
        BP2_norm, BP2 = beam_pattern(dir_s0=dir_s20, dir = dir_s2, sigma=sigma_l, I_0=I_0, I_hat = I_hat)

        # Returns an n by 1 vector of cosines
        gamma_s1 = compute_gamma(dir = dir_s1, normals = norm_vec)
        gamma_s2 = compute_gamma(dir=dir_s2, normals = norm_vec)

        # Simple computation of incident light
        E_i1 = BP1 * np.cos(gamma_s1) * np.exp(-k * R_s1) / (R_s1 ** 2)
        E_i2 = BP2 * np.cos(gamma_s2) * np.exp(-k * R_s2) / (R_s2 ** 2)






        if is_skogn == False:
            E_i = E_i1 + E_i2
        else:
            E_i = E_i2



        # Calculate reflected light:
        L_r = (rho_gt/np.pi)*E_i

        L_d = L_r*np.exp(-k*R_c)

        # Compute backscatter
        include_backscatter = False
        if include_backscatter:
            L_s1 = compute_scattering_integral(k, R_s1)
            L_s2 = compute_scattering_integral(k, R_s2)
        else:
            L_s1 = 0
            L_s2 = 0

        L_s = L_s1 + L_s2

        L_w = L_d + L_s

        # Calculate from measurement to seafloor
        theta_w = theta_w_dot - alpha  # Intersection angle with respect to glass

        I_f = immersion_factor(theta_w=theta_w, n_w=n_w, n_g=n_g, pixel_nr=pixel_nr)

        L_air_est = np.einsum('ij,ij -> ij', L_w, 1/I_f)

        rad_2_DC = radiance_conversion(RF = RF, t_exp = t_exp, pixel_nr = pixel_nr)

        DC_air_est = np.einsum('ij,ij -> ij', L_air_est, rad_2_DC)

        DC_air_meas = np.einsum('ij,ij -> ij', L_air, rad_2_DC)

        DC_diff = DC_air_meas - DC_air_est

        return DC_diff
    #    # The measurement is L_a meaning radiance in air
#
#
    #    L_m = L_plus*I_f
#


